from regressions.errors.scalars import scalar_value
from regressions.errors.matrices import matrix_of_scalars, columns_rows
from regressions.vectors.multiplication import dot_product
from .transpose import transposed_matrix

def scalar_product_matrix(matrix, scalar):
    """
    Calculates the product of a matrix and a scalar

    Parameters
    ----------
    matrix : list of lists of int or float
        List of lists of numbers representing a matrix
    scalar : int or float
        Number representing a scalar

    Raises
    ------
    TypeError
        First argument must be 2-dimensional lists
    TypeError
        Elements nested within first argument must be integers or floats
    TypeError
        Second argument must be an integer or a float

    Returns
    -------
    matrix : list of lists of int or float
        List of lists in which each inner element is the product of the corresponding element from the input matrix and the scalar value

    See Also
    --------
    :func:`~regressions.vectors.multiplication.scalar_product_vector`, :func:`~regressions.matrices.addition.matrix_sum`

    Notes
    -----
    - Matrix: :math:`\\mathbf{A} = \\begin{bmatrix} a_{1,1} & a_{1,2} & \\cdots & a_{1,n} \\\\ a_{2,1} & a_{2,2} & \\cdots & a_{2,n} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ a_{m,1} & a_{m,2} & \\cdots & a_{m,n} \\end{bmatrix}`
    - Scalar: :math:`c`
    - Scalar product: :math:`c\\cdot{\\mathbf{A}} = \\begin{bmatrix} c\\cdot{a_{1,1}} & c\\cdot{a_{1,2}} & \\cdots & c\\cdot{a_{1,n}} \\\\ c\\cdot{a_{2,1}} & c\\cdot{a_{2,2}} & \\cdots & c\\cdot{a_{2,n}} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ c\\cdot{a_{m,1}} & c\\cdot{a_{m,2}} & \\cdots & c\\cdot{a_{m,n}} \\end{bmatrix}`
    - |matrix_scalar_multiplication|

    Examples
    --------
    Import `scalar_product_matrix` function from `regressions` library
        >>> from regressions.matrices.multiplication import scalar_product_matrix
    Multiply [[1, 2, 3], [4, 5, 6]] and -2
        >>> matrix_2x3 = scalar_product_matrix([[1, 2, 3], [4, 5, 6]], -2)
        >>> print(matrix_2x3)
        [[-2, -4, -6], [-8, -10, -12]]
    Multiply [[5, -7], [-3, 8]] and 3
        >>> matrix_2x2 = scalar_product_matrix([[5, -7], [-3, 8]], 3)
        >>> print(matrix_2x2)
        [[15, -21], [-9, 24]]
    """
    # Handle input errors
    matrix_of_scalars(matrix, 'first')
    scalar_value(scalar, 'second')

    # Create list to return
    result = []

    # Iterate over outer lists of input
    for m in range(len(matrix)):
        # Create new lists inside list to return
        result.append([])

        # Iterate over inner lists of input
        for n in range(len(matrix[0])):
            # Store products in inner lists of return
            result[m].append(matrix[m][n] * scalar)
    
    # Return result
    return result

def matrix_product(matrix_one, matrix_two):
    """
    Calculates the product of two matrices

    Parameters
    ----------
    matrix_one : list of lists of int or float
        List of lists of numbers representing a matrix
    matrix_two : list of lists of int or float
        List of lists of numbers representing a matrix

    Raises
    ------
    TypeError
        Arguments must be 2-dimensional lists
    TypeError
        Elements nested within arguments must be integers or floats
    ValueError
        First list within first argument must contain the same amount of elements as the amount of lists contained within second argument

    Returns
    -------
    matrix : list of lists of float
        List of lists in which each inner element is the dot product of the first matrix's row vector corresponding to that element's row position and the second matrix's column vector corresponding to that element's column position; resultant matrix will have the same number of rows as the first matrix and the same number of columns as the second matrix

    See Also
    --------
    :func:`~regressions.vectors.multiplication.dot_product`, :func:`~regressions.matrices.transpose.transposed_matrix`

    Notes
    -----
    - First matrix: :math:`\\mathbf{A} = \\begin{bmatrix} a_{1,1} & a_{1,2} & \\cdots & a_{1,n} \\\\ a_{2,1} & a_{2,2} & \\cdots & a_{2,n} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ a_{m,1} & a_{m,2} & \\cdots & a_{m,n} \\end{bmatrix}`
    - Second matrix: :math:`\\mathbf{B} = \\begin{bmatrix} b_{1,1} & b_{1,2} & \\cdots & b_{1,p} \\\\ b_{2,1} & b_{2,2} & \\cdots & b_{2,p} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ b_{n,1} & b_{n,2} & \\cdots & b_{n,p} \\end{bmatrix}`
    - Product of matrices: :math:`\\mathbf{A}\\cdot{\\mathbf{B}} = \\begin{bmatrix} a_{1,1}\\cdot{b_{1,1}} + a_{1,2}\\cdot{b_{2,1}} + \\cdots + a_{1,n}\\cdot{b_{n,1}} & a_{1,1}\\cdot{b_{1,2}} + a_{1,2}\\cdot{b_{2,2}} + \\cdots + a_{1,n}\\cdot{b_{n,2}} & \\cdots & a_{1,1}\\cdot{b_{1,p}} + a_{1,2}\\cdot{b_{2,p}} + \\cdots + a_{1,n}\\cdot{b_{n,p}} \\\\ a_{2,1}\\cdot{b_{1,1}} + a_{2,2}\\cdot{b_{2,1}} + \\cdots + a_{2,n}\\cdot{b_{n,1}} & a_{2,1}\\cdot{b_{1,2}} + a_{2,2}\\cdot{b_{2,2}} + \\cdots + a_{2,n}\\cdot{b_{n,2}} & \\cdots & a_{2,1}\\cdot{b_{1,p}} + a_{2,2}\\cdot{b_{2,p}} + \\cdots + a_{2,n}\\cdot{b_{n,p}} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ a_{m,1}\\cdot{b_{1,1}} + a_{m,2}\\cdot{b_{2,1}} + \\cdots + a_{m,n}\\cdot{b_{n,1}} & a_{m,1}\\cdot{b_{1,2}} + a_{m,2}\\cdot{b_{2,2}} + \\cdots + a_{m,n}\\cdot{b_{n,2}} & \\cdots & a_{m,1}\\cdot{b_{1,p}} + a_{m,2}\\cdot{b_{2,p}} + \\cdots + a_{m,n}\\cdot{b_{n,p}} \\end{bmatrix}`
    - |matrix_multiplication|

    Examples
    --------
    Import `matrix_product` function from `regressions` library
        >>> from regressions.matrices.multiplication import matrix_product
    Multiply [[1, 2, 3], [4, 5, 6]] and [[2, 3], [5, 7], [11, 13]]
        >>> matrix_2x2 = matrix_product([[1, 2, 3], [4, 5, 6]], [[2, 3], [5, 7], [11, 13]])
        >>> print(matrix_2x2)
        [[45.0, 56.0], [99.0, 125.0]]
    Multiply [[1, 2, 3], [4, 5, 6]] and [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]]
        >>> matrix_2x4 = matrix_product([[1, 2, 3], [4, 5, 6]], [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]])
        >>> print(matrix_2x4)
        [[38.0, 44.0, 50.0, 56.0], [83.0, 98.0, 113.0, 128.0]]
    """
    # Handle input errors
    columns_rows(matrix_one, matrix_two)

    # Create list to return
    result = []

    # Iterate over outer lists of input
    for m in range(len(matrix_one)):
        # Create new lists inside list to return
        result.append([])

        # Iterate over inner lists of input
        for n in range(len(matrix_two[0])):
            # Store dot products in inner lists of return
            result[m].append(dot_product(matrix_one[m], transposed_matrix(matrix_two)[n]))
    
    # Return result
    return result


from math import log
from regressions.errors.scalars import two_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value, rounded_list

def exponential_derivatives(first_constant, second_constant, precision = 4):
    """
    Calculates the first and second derivatives of an exponential function

    Parameters
    ----------
    first_constant : int or float
        Constant multiple of the original exponential function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Base rate of variable of the original exponential function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First two arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    derivatives['first']['constants'] : list of float
        Coefficients of the resultant first derivative
    derivatives['first']['evaluation'] : func
        Function for evaluating the resultant first derivative at any float or integer argument
    derivatives['second']['constants'] : list of float
        Coefficients of the resultant second derivative
    derivatives['second']['evaluation'] : func
        Function for evaluating the resultant second derivative at any float or integer argument

    See Also
    --------
    :func:`~regressions.analyses.equations.exponential.exponential_equation`, :func:`~regressions.analyses.integrals.exponential.exponential_integral`, :func:`~regressions.analyses.roots.exponential.exponential_roots`, :func:`~regressions.models.exponential.exponential_model`

    Notes
    -----
    - Standard form of an exponential function: :math:`f(x) = a\\cdot{b^x}`
    - First derivative of an exponential function: :math:`f'(x) = a\\cdot{\\ln{b}\\cdot{b^x}}`
    - Second derivative of an exponential function: :math:`f''(x) = a\\cdot{\\ln^2{b}\\cdot{b^x}}`
    - |differentiation_formulas|
    - |exponential|

    Examples
    --------
    Import `exponential_derivatives` function from `regressions` library
        >>> from regressions.analyses.derivatives.exponential import exponential_derivatives
    Generate the derivatives of an exponential function with coefficients 2 and 3, then display the coefficients of its first and second derivatives
        >>> derivatives_constants = exponential_derivatives(2, 3)
        >>> print(derivatives_constants['first']['constants'])
        [2.1972, 3.0]
        >>> print(derivatives_constants['second']['constants'])
        [2.4139, 3.0]
    Generate the derivatives of an exponential function with coefficients -2 and 3, then evaluate its first and second derivatives at 10
        >>> derivatives_evaluation = exponential_derivatives(-2, 3)
        >>> print(derivatives_evaluation['first']['evaluation'](10))
        -129742.4628
        >>> print(derivatives_evaluation['second']['evaluation'](10))
        -142538.3811
    Generate the derivatives of an exponential function with all inputs set to 0, then display the coefficients of its first and second derivatives
        >>> derivatives_zeroes = exponential_derivatives(0, 0)
        >>> print(derivatives_zeroes['first']['constants'])
        [-0.0009, 0.0001]
        >>> print(derivatives_zeroes['second']['constants'])
        [0.0083, 0.0001]
    """
    # Handle input errors
    two_scalars(first_constant, second_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant], precision)

    # Create first derivative
    first_coefficients = [coefficients[0] * log(abs(coefficients[1])), coefficients[1]]
    first_constants = rounded_list(first_coefficients, precision)
    def first_derivative(variable):
        evaluation = first_constants[0] * first_constants[1]**variable
        rounded_evaluation = rounded_value(evaluation, precision)
        return rounded_evaluation
    first_dictionary = {
        'constants': first_constants,
        'evaluation': first_derivative
    }

    # Create second derivative
    second_coefficients = [first_constants[0] * log(abs(first_constants[1])), first_constants[1]]
    second_constants = rounded_list(second_coefficients, precision)
    def second_derivative(variable):
        evaluation = second_constants[0] * second_constants[1]**variable
        rounded_evaluation = rounded_value(evaluation, precision)
        return rounded_evaluation
    second_dictionary = {
        'constants': second_constants,
        'evaluation': second_derivative
    }

    # Package both derivatives in single dictionary
    results = {
        'first': first_dictionary,
        'second': second_dictionary
    }
    return results


from regressions.errors.vectors import compare_vectors
from regressions.errors.scalars import positive_integer
from .residuals import multiple_residuals
from .deviations import multiple_deviations
from .summation import sum_value
from .rounding import rounded_value

def correlation_coefficient(actuals, expecteds, precision = 4):
    """
    Calculates the correlation coefficient as a way to predict the strength of a predicted model by comparing the ratio of residuals to deviations, in order to determine a strong or weak relationship

    Parameters
    ----------
    actuals : list of int or float
        List containing the actual values observed from a data set
    expecteds : list of int or float
        List containing the expected values for a data set based on a predictive model
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the result

    Raises
    ------
    TypeError
        First and second arguments must be 1-dimensional lists
    TypeError
        Elements of first and second arguments must be integers or floats
    ValueError
        First and second arguments must contain the same number of elements
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    correlation : float
        Number indicating statistical strenght of the relationship between two variables; the closer to 1, the stronger; the closer to 0, the weaker

    See Also
    --------
    :func:`~regressions.statistics.residuals.multiple_residuals`, :func:`~regressions.statistics.deviations.multiple_deviations`, :func:`~regressions.statistics.summation.sum_value`

    Notes
    -----
    - Observed values: :math:`y_i = \\{ y_1, y_2, \\cdots, y_n \\}`
    - Predicted values: :math:`\\hat{y}_i = \\{ \\hat{y}_1, \\hat{y}_2, \\cdots, \\hat{y}_n \\}`
    - Mean of all observed values: :math:`\\bar{y} = \\frac{1}{n}\\cdot{\\sum\\limits_{i=1}^n y_i}`
    - Residuals: :math:`e_i = \\{ y_1 - \\hat{y}_1, y_2 - \\hat{y}_2, \\cdots, y_n - \\hat{y}_n \\}`
    - Deviations: :math:`d_i = \\{ y_1 - \\bar{y}, y_2 - \\bar{y}, \\cdots, y_n - \\bar{y} \\}`
    - Sum of squares of residuals: :math:`SS_{res} = \\sum\\limits_{i=1}^n e_i^2`
    - Sum of squares of deviations: :math:`SS_{dev} = \\sum\\limits_{i=1}^n d_i^2`
    - Correlation coefficient: :math:`r = \\sqrt{1 - \\frac{SS_{res}}{SS_{dev}}}`
    - |determination|

    Examples
    --------
    Import `correlation_coefficient` function from `regressions` library
        >>> from regressions.statistics.correlation import correlation_coefficient
    Calculate the correlation using the provided actual values [8.2, 9.41, 1.23, 34.7] and the predicted values [7.863, 8.9173, 2.0114, 35.8021]
        >>> correlation_short = correlation_coefficient([8.2, 9.41, 1.23, 34.7], [7.863, 8.9173, 2.0114, 35.8021])
        >>> print(correlation_short)
        0.9983
    Calculate the correlation using the provided actual values [2, 3, 5, 7, 11, 13, 17, 19] and the predicted values [1.0245, 3.7157, 6.1398, 8.1199, 12.7518, 14.9621, 15.2912, 25.3182]
        >>> correlation_long = correlation_coefficient([2, 3, 5, 7, 11, 13, 17, 19], [1.0245, 3.7157, 6.1398, 8.1199, 12.7518, 14.9621, 15.2912, 25.3182])
        >>> print(correlation_long)
        0.9011
    """
    # Handle input errors
    compare_vectors(actuals, expecteds)
    positive_integer(precision)

    # Determine residuals and deviations between actuals and expecteds
    residuals = multiple_residuals(actuals, expecteds)
    deviations = multiple_deviations(actuals)

    # Square residuals and deviations
    squared_residuals = []
    for residual in residuals:
        squared_residuals.append(residual**2)
    squared_deviations = []
    for deviation in deviations:
        squared_deviations.append(deviation**2)
    
    # Calculate the sums of the squares of the residuals and deviations
    residual_sum = sum_value(squared_residuals)
    deviation_sum = sum_value(squared_deviations)

    # Circumvent division by zero
    if deviation_sum == 0:
        deviation_sum = 10**(-precision)
    
    # Calculate key ratio
    ratio = residual_sum / deviation_sum
    
    # Handle no solution
    if ratio > 1:
        return 0.0
    
    # Handle general case
    else:
        result = (1 - ratio)**(1/2)
        return rounded_value(result, precision)


from regressions.errors.scalars import select_integers, positive_integer
from regressions.errors.vectors import vector_of_scalars
from regressions.errors.analyses import select_equations
from .roots.linear import linear_roots_first_derivative, linear_roots_second_derivative
from .roots.quadratic import quadratic_roots_first_derivative, quadratic_roots_second_derivative
from .roots.cubic import cubic_roots_first_derivative, cubic_roots_second_derivative
from .roots.hyperbolic import hyperbolic_roots_first_derivative, hyperbolic_roots_second_derivative
from .roots.exponential import exponential_roots_first_derivative, exponential_roots_second_derivative
from .roots.logarithmic import logarithmic_roots_first_derivative, logarithmic_roots_second_derivative
from .roots.logistic import logistic_roots_first_derivative, logistic_roots_second_derivative
from .roots.sinusoidal import sinusoidal_roots_first_derivative, sinusoidal_roots_second_derivative

def critical_points(equation_type, coefficients, derivative_level, precision = 4):
    """
    Calculates the critical points of a specific function at a certain derivative level

    Parameters
    ----------
    equation_type : str
        Name of the type of function for which critical points must be determined (e.g., 'linear', 'quadratic')
    coefficients : list of int or float
        Coefficients to use to generate the equation to investigate
    derivative_level : int
        Integer corresponding to which derivative to investigate for critical points (1 for the first derivative and 2 for the second derivative)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the results

    Raises
    ------
    ValueError
        First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'
    TypeError
        Second argument must be a 1-dimensional list containing elements that are integers or floats
    ValueError
        Third argument must be one of the following integers: [1, 2]
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    points : list of float or str
        Values of the x-coordinates at which the original function's derivative either crosses the x-axis or does not exist; if the function is sinusoidal, then only five results within a two-period interval will be listed, but a general form will also be included; if the derivative has no critical points, then it will return a list of `None`

    See Also
    --------
    - Roots for key functions: :func:`~regressions.analyses.roots.linear.linear_roots`, :func:`~regressions.analyses.roots.quadratic.quadratic_roots`, :func:`~regressions.analyses.roots.cubic.cubic_roots`, :func:`~regressions.analyses.roots.hyperbolic.hyperbolic_roots`, :func:`~regressions.analyses.roots.exponential.exponential_roots`, :func:`~regressions.analyses.roots.logarithmic.logarithmic_roots`, :func:`~regressions.analyses.roots.logistic.logistic_roots`, :func:`~regressions.analyses.roots.sinusoidal.sinusoidal_roots`
    - Graphical analysis: :func:`~regressions.analyses.intervals.sign_chart`, :func:`~regressions.analyses.points.key_coordinates`

    Notes
    -----
    - Domain of a function: :math:`x_i = \\{ x_1, x_2, \\cdots, x_n \\}`
    - Potential critical points of the derivative of the function: :math:`x_c = \\{ c \\mid c \\in x_i, f'(c) = 0 \\cup f'(c) = \\varnothing \\}`
    - |critical_points|

    Examples
    --------
    Import `critical_points` function from `regressions` library
        >>> from regressions.analyses.criticals import critical_points
    Calulate the critical points of the second derivative of a cubic function with coefficients 2, 3, 5, and 7
        >>> points_cubic = critical_points('cubic', [2, 3, 5, 7], 2)
        >>> print(points_cubic)
        [-0.5]
    Calulate the critical points of the first derivative of a sinusoidal function with coefficients 2, 3, 5, and 7
        >>> points_sinusoidal = critical_points('sinusoidal', [2, 3, 5, 7], 1)
        >>> print(points_sinusoidal)
        [5.5236, 6.5708, 7.618, 8.6652, 9.7124, '5.5236 + 1.0472k']
    """
    # Handle input errors
    select_equations(equation_type)
    vector_of_scalars(coefficients, 'second')
    select_integers(derivative_level, [1, 2], 'third')
    positive_integer(precision)

    # Create list to return
    results = []

    # Determine critical points for first derivative based on equation type
    if derivative_level == 1:
        if equation_type == 'linear':
            results = linear_roots_first_derivative(*coefficients, precision)
        elif equation_type == 'quadratic':
            results = quadratic_roots_first_derivative(*coefficients, precision)
        elif equation_type == 'cubic':
            results = cubic_roots_first_derivative(*coefficients, precision)
        elif equation_type == 'hyperbolic':
            results = hyperbolic_roots_first_derivative(*coefficients, precision)
        elif equation_type == 'exponential':
            results = exponential_roots_first_derivative(*coefficients, precision)
        elif equation_type == 'logarithmic':
            results = logarithmic_roots_first_derivative(*coefficients, precision)
        elif equation_type == 'logistic':
            results = logistic_roots_first_derivative(*coefficients, precision)
        elif equation_type == 'sinusoidal':
            results = sinusoidal_roots_first_derivative(*coefficients, precision)

    # Determine critical points for second derivative based on equation type
    elif derivative_level == 2:
        if equation_type == 'linear':
            results = linear_roots_second_derivative(*coefficients, precision)
        elif equation_type == 'quadratic':
            results = quadratic_roots_second_derivative(*coefficients, precision)
        elif equation_type == 'cubic':
            results = cubic_roots_second_derivative(*coefficients, precision)
        elif equation_type == 'hyperbolic':
            results = hyperbolic_roots_second_derivative(*coefficients, precision)
        elif equation_type == 'exponential':
            results = exponential_roots_second_derivative(*coefficients, precision)
        elif equation_type == 'logarithmic':
            results = logarithmic_roots_second_derivative(*coefficients, precision)
        elif equation_type == 'logistic':
            results = logistic_roots_second_derivative(*coefficients, precision)
        elif equation_type == 'sinusoidal':
            results = sinusoidal_roots_second_derivative(*coefficients, precision)
    return results


from math import asin, acos, pi
from regressions.errors.scalars import four_scalars, five_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.sort import sorted_list, sorted_strings
from regressions.statistics.rounding import rounded_value, rounded_list
from regressions.vectors.generate import generate_elements
from regressions.vectors.separate import separate_elements
from regressions.analyses.derivatives.sinusoidal import sinusoidal_derivatives

def sinusoidal_roots(first_constant, second_constant, third_constant, fourth_constant, precision = 4):
    """
    Calculates the roots of a sinusoidal function

    Parameters
    ----------
    first_constant : int or float
        Vertical stretch factor of the original sine function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Horizontal stretch factor of the original sine function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    third_constant : int or float
        Horizontal shift of the original sine function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    fourth_constant : int or float
        Vertical shift of the original sine function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First four arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    roots : list of float or str
        List of the x-coordinates of the initial x-intercepts within two periods of the original function in float format, along with the general forms in string format that can be used to determine all other x-intercepts by plugging in any integer value for 'k' and evaluating; if the function never crosses the x-axis, then it will return a list of `None`

    See Also
    --------
    :func:`~regressions.analyses.equations.sinusoidal.sinusoidal_equation`, :func:`~regressions.analyses.derivatives.sinusoidal.sinusoidal_derivatives`, :func:`~regressions.analyses.integrals.sinusoidal.sinusoidal_integral`, :func:`~regressions.models.sinusoidal.sinusoidal_model`

    Notes
    -----
    - Standard form of a sinusoidal function: :math:`f(x) = a\\cdot{\\sin(b\\cdot(x - c))} + d`
    - Sinusoidal formula: :math:`x_0 = c + \\frac{1}{b}\\cdot{\\sin^{-1}(-\\frac{d}{a})} + \\frac{2\\pi}{b}\\cdot{k}`

        - :math:`\\text{if} -1 < -\\frac{d}{a} < 0 \\text{ or } 0 < -\\frac{d}{a} < 1, x_1 = c + \\frac{\\pi}{b} - \\frac{1}{b}\\cdot{\\sin^{-1}(-\\frac{d}{a})} + \\frac{2\\pi}{b}\\cdot{k}`
        - :math:`\\text{if} -\\frac{d}{a} = 0, x_1 = c - \\frac{\\pi}{b} + \\frac{2\\pi}{b}\\cdot{k}`
        - :math:`k \\in \\mathbb{Z}`

    Examples
    --------
    Import `sinusoidal_roots` function from `regressions` library
        >>> from regressions.analyses.roots.sinusoidal import sinusoidal_roots
    Calculate the roots of a sinusoidal function with coefficients 2, 3, 5, and 7
        >>> roots_first = sinusoidal_roots(2, 3, 5, 7)
        >>> print(roots_first)
        [None]
    Calculate the roots of a sinusoidal function with coefficients 7, -5, -3, and 2
        >>> roots_second = sinusoidal_roots(7, -5, -3, 2)
        >>> print(roots_second)
        [-8.7128, -7.9686, -7.4562, -6.712, -6.1995, -5.4553, -4.9429, -4.1987, -3.6863, -2.942, '-3.6863 + -1.2566k', '-2.942 + -1.2566k']
    Calculate the roots of a sinusoidal function with all inputs set to 0
        >>> roots_zeroes = sinusoidal_roots(0, 0, 0, 0)
        >>> print(roots_zeroes)
        [-15707.9632, 47123.8899, 109955.743, 172787.596, 235619.4491, '-15707.9632 + 62831.8531k']
    """
    # Handle input errors
    four_scalars(first_constant, second_constant, third_constant, fourth_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant, third_constant, fourth_constant], precision)

    # Create list for roots
    roots = []

    # Identify key ratio
    ratio = -1 * coefficients[3] / coefficients[0]

    # Handle no roots
    if ratio > 1 or ratio < -1:
        roots = [None]
    
    # Handle multiple roots
    else:
        # Create intermediary variables
        radians = asin(ratio)
        periodic_radians = radians / coefficients[1]

        # Determine pertinent values
        periodic_unit = 2 * pi / coefficients[1]
        initial_value = coefficients[2] + periodic_radians
        roots = generate_elements(initial_value, periodic_unit, precision)
        
        # Handle roots that bounce on the x-axis
        if ratio == 1 or ratio == -1:
            pass
        
        # Handle roots that cross the x-axis
        else:
            # Determine supplementary values
            alternative_initial_value = coefficients[2] + pi / coefficients[1] - periodic_radians
            generated_elements = generate_elements(alternative_initial_value, periodic_unit, precision)
            
            # Add additional results to roots list
            roots.extend(generated_elements)
    
    # Separate numerical roots, string roots, and None results
    separated_roots = separate_elements(roots)
    numerical_roots = separated_roots['numerical']
    other_roots = separated_roots['other']
    
    # Sort numerical roots
    sorted_roots = sorted_list(numerical_roots)

    # Round numerical roots
    rounded_roots = rounded_list(sorted_roots, precision)
    
    # Sort other_roots
    sorted_other_roots = sorted_strings(other_roots)

    # Combine numerical and non-numerical roots
    result = rounded_roots + sorted_other_roots
    return result

def sinusoidal_roots_first_derivative(first_constant, second_constant, third_constant, fourth_constant, precision = 4):
    # Handle input errors
    four_scalars(first_constant, second_constant, third_constant, fourth_constant)
    positive_integer(precision)

    # Generate coefficients of first derivative
    constants = sinusoidal_derivatives(first_constant, second_constant, third_constant, fourth_constant)['first']['constants']

    # Create intermediary variables
    periodic_unit = pi / constants[1]
    initial_value = constants[2] + pi / (2 * constants[1])

    # Generate roots based on criteria
    generated_roots = generate_elements(initial_value, periodic_unit, precision)

    # Return result
    result = generated_roots
    return result

def sinusoidal_roots_second_derivative(first_constant, second_constant, third_constant, fourth_constant, precision = 4):
    # Handle input errors
    four_scalars(first_constant, second_constant, third_constant, fourth_constant)
    positive_integer(precision)

    # Generate coefficients of second derivative
    constants = sinusoidal_derivatives(first_constant, second_constant, third_constant, fourth_constant)['second']['constants']

    # Create intermediary variables
    periodic_unit = pi / constants[1]
    initial_value = constants[2]

    # Generate roots based on criteria
    generated_roots = generate_elements(initial_value, periodic_unit, precision)

    # Return result
    result = generated_roots
    return result

def sinusoidal_roots_initial_value(first_constant, second_constant, third_constant, fourth_constant, initial_value, precision = 4):
    # Handle input errors
    five_scalars(first_constant, second_constant, third_constant, fourth_constant, initial_value)
    positive_integer(precision)

    # Determine roots given an initial value
    result = sinusoidal_roots(first_constant, second_constant, third_constant, fourth_constant - initial_value, precision)
    return result

def sinusoidal_roots_derivative_initial_value(first_constant, second_constant, third_constant, fourth_constant, initial_value, precision = 4):
    # Handle input errors
    five_scalars(first_constant, second_constant, third_constant, fourth_constant, initial_value)
    positive_integer(precision)

    # Create intermediary list and list to return
    roots = []
    result = []

    # Identify key ratio
    ratio = initial_value / (first_constant * second_constant)

    # Handle no roots
    if ratio > 1 or ratio < -1:
        result.append(None)
    
    # Handle multiple roots
    else:
        # Handle case in which initial value is zero
        if ratio == 0:
            roots = sinusoidal_roots_first_derivative(first_constant, second_constant, third_constant, fourth_constant, precision)
        
        # Handle general case
        else:
            radians = acos(ratio)
            periodic_radians = radians / second_constant
            periodic_unit = 2 * pi / second_constant
            initial = third_constant + periodic_radians
            roots = generate_elements(initial, periodic_unit, precision)

            # Handle roots that bounce on the x-axis
            if ratio == 1 or ratio == -1:
                pass

            # Handle roots that cross the x-axis
            else:
                alternative_initial = third_constant + periodic_unit - periodic_radians
                generated_elements = generate_elements(alternative_initial, periodic_unit, precision)
                roots.extend(generated_elements)
        
        # Separate numerical roots, string roots, and None results
        separated_roots = separate_elements(roots)
        numerical_roots = separated_roots['numerical']
        other_roots = separated_roots['other']
        
        # Sort numerical roots
        sorted_roots = sorted_list(numerical_roots)

        # Round numerical roots
        rounded_roots = rounded_list(sorted_roots, precision)
        
        # Sort other_roots
        sorted_other_roots = sorted_strings(other_roots)

        # Combine numerical and non-numerical roots
        result.extend(rounded_roots + sorted_other_roots)
    
    # Return result
    return result


from math import log
from regressions.errors.scalars import two_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value

def logarithmic_integral(first_constant, second_constant, precision = 4):
    """
    Generates the integral of a logarithmic function

    Parameters
    ----------
    first_constant : int or float
        Coefficient of the logarithmic term of the original logarithmic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Coefficient of the constant term of the original logarithmic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First two arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    integral['constants'] : list of float
        Coefficients of the resultant integral
    integral['evaluation'] : func
        Function for evaluating the resultant integral at any float or integer argument; if zero inputted as argument, it will be converted to a small, non-zero decimal value (e.g., 0.0001)

    See Also
    --------
    :func:`~regressions.analyses.equations.logarithmic.logarithmic_equation`, :func:`~regressions.analyses.derivatives.logarithmic.logarithmic_derivatives`, :func:`~regressions.analyses.roots.logarithmic.logarithmic_roots`, :func:`~regressions.models.logarithmic.logarithmic_model`

    Notes
    -----
    - Standard form of a logarithmic function: :math:`f(x) = a\\cdot{\\ln{x}} + b`
    - Integral of a logarithmic function: :math:`F(x) = a\\cdot{x}\\cdot(\\ln{x} - 1) + b\\cdot{x}`
    - |indefinite_integral|
    - |integration_formulas|
    - |integration_by_parts|

    Examples
    --------
    Import `logarithmic_integral` function from `regressions` library
        >>> from regressions.analyses.integrals.logarithmic import logarithmic_integral
    Generate the integral of a logarithmic function with coefficients 2 and 3, then display its coefficients
        >>> integral_constants = logarithmic_integral(2, 3)
        >>> print(integral_constants['constants'])
        [2.0, 3.0]
    Generate the integral of a logarithmic function with coefficients -2 and 3, then evaluate its integral at 10
        >>> integral_evaluation = logarithmic_integral(-2, 3)
        >>> print(integral_evaluation['evaluation'](10))
        3.9483
    Generate the integral of a logarithmic function with all inputs set to 0, then display its coefficients
        >>> integral_zeroes = logarithmic_integral(0, 0)
        >>> print(integral_zeroes['constants'])
        [0.0001, 0.0001]
    """
    # Handle input errors
    two_scalars(first_constant, second_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant], precision)

    # Create constants
    constants = [coefficients[0], coefficients[1]]

    # Create evaluation
    def logarithmic_evaluation(variable):
        # Circumvent logarithm of zero
        if variable == 0:
            variable = 10**(-precision)
        evaluation = constants[0] * variable * (log(abs(variable)) - 1) + constants[1] * variable
        rounded_evaluation = rounded_value(evaluation, precision)
        return rounded_evaluation
    
    # Package constants and evaluation in single dictionary
    results = {
        'constants': constants,
        'evaluation': logarithmic_evaluation
    }
    return results


from math import floor
from regressions.errors.vectors import vector_of_scalars
from .sort import sorted_list

def median_value(data):
    """
    Determines the median value of a data set

    Parameters
    ----------
    data : list of int or float
        List of numbers to analyze

    Raises
    ------
    TypeError
        Argument must be a 1-dimensional list
    TypeError
        Elements of argument must be integers or floats

    Returns
    -------
    median : int or float
        Middle value of the data set, splitting the data evenly in half

    See Also
    --------
    :func:`~regressions.statistics.sort.sorted_list`, :func:`~regressions.statistics.minimum.minimum_value`, :func:`~regressions.statistics.maximum.maximum_value`

    Notes
    -----
    - Ordered set of numbers: :math:`a_i = ( a_1, a_2, \\cdots, a_n )`
    - Median of all numbers in set (if set contains an odd amount of numbers): :math:`M = a_{\\lceil n/2 \\rceil}`
    - Median of all numbers in set (if set contains an even amount of numbers): :math:`M = \\frac{a_{n/2} + a_{n/2 + 1}}{2}`
    - |median|

    Examples
    --------
    Import `median_value` function from `regressions` library
        >>> from regressions.statistics.median import median_value
    Determine the median of the set [21, 53, 3, 68, 43, 9, 72, 19, 20, 1]
        >>> median_even = median_value([21, 53, 3, 68, 43, 9, 72, 19, 20, 1])
        >>> print(median_even)
        20.5
    Determine the median of the set [12, 81, 13, 8, 42, 72, 91, 20, 20]
        >>> median_odd = median_value([12, 81, 13, 8, 42, 72, 91, 20, 20])
        >>> print(median_odd)
        20
    """
    # Handle input errors
    vector_of_scalars(data)

    # Sort input
    sorted_data = sorted_list(data)

    # Determine length of input
    length = len(sorted_data)

    # Handle an even amount of elements
    if length % 2 == 0:
        upper_index = int(length / 2)
        lower_index = int(upper_index - 1)
        upper_value = sorted_data[upper_index]
        lower_value = sorted_data[lower_index]
        result = (upper_value + lower_value) / 2
        return result
    
    # Handle an odd amount of elements
    else:
        index = int(floor(length / 2))
        result = sorted_data[index]
        return result


from regressions.errors.matrices import matrix_of_scalars

def transposed_matrix(matrix):
    """
    Transpose a matrix's rows and columns

    Parameters
    ----------
    matrix : list of lists of int or float
        List of lists of numbers representing a matrix

    Raises
    ------
    TypeError
        Argument must be a 2-dimensional list
    TypeError
        Elements nested within argument must be integers or floats
    
    Returns
    -------
    matrix : list of lists of int or float
        List of lists in which each inner element occupies the row that correspond's to the column it occupied in the original matrix and the column that correspond's to the row it occupied in the original matrix

    See Also
    --------
    :func:`~regressions.matrices.cofactors.matrix_of_cofactors`, :func:`~regressions.matrices.minors.matrix_of_minors`, :func:`~regressions.matrices.determinant.linear_determinant`, :func:`~regressions.matrices.inverse.inverse_matrix`

    Notes
    -----
    - Original matrix: :math:`\\mathbf{A} = \\begin{bmatrix} a_{1,1} & a_{1,2} & \\cdots & a_{1,n} \\\\ a_{2,1} & a_{2,2} & \\cdots & a_{2,n} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ a_{m,1} & a_{m,2} & \\cdots & a_{m,n} \\end{bmatrix}`
    - Transpose of matrix: :math:`\\mathbf{A}^T = \\begin{bmatrix} a_{1,1} & a_{2,1} & \\cdots & a_{m,1} \\\\ a_{1,2} & a_{2,2} & \\cdots & a_{m,2} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ a_{1,n} & a_{2,n} & \\cdots & a_{m,n} \\end{bmatrix}`
    - |adjugate|

    Examples
    --------
    Import `transposed_matrix` function from `regressions` library
        >>> from regressions.matrices.transpose import transposed_matrix
    Transpose [[1, 2, 3], [4, 5, 6]]
        >>> matrix_3x2 = transposed_matrix([[1, 2, 3], [4, 5, 6]])
        >>> print(matrix_3x2)
        [[1, 4], [2, 5], [3, 6]]
    Transpose [[2, 3], [5, 7]]
        >>> matrix_2x2 = transposed_matrix([[2, 3], [5, 7]])
        >>> print(matrix_2x2)
        [[2, 5], [3, 7]]
    """
    # Handle input errors
    matrix_of_scalars(matrix)

    # Create list to return
    result = []

    # Iterate over inner lists of input
    for m in range(len(matrix[0])):
        # Create new lists inside list to return
        result.append([])

        # Iterate over outer lists of input
        for n in range(len(matrix)):
            # Store elements previously at inner-outer, now at outer-inner
            result[m].append(matrix[n][m])
    
    # Return result
    return result


from regressions.statistics.rounding import rounded_value

def no_zeroes(coefficients, precision = 4):
    result = []
    for coefficient in coefficients:
        if coefficient == 0:
            result.append(10**(-precision))
        else:
            result.append(rounded_value(coefficient, precision))
    return result


from regressions.errors.vectors import vector_of_scalars
from .sort import sorted_list

def maximum_value(data):
    """
    Determines the largest value of a data set

    Parameters
    ----------
    data : list of int or float
        List of numbers to analyze

    Raises
    ------
    TypeError
        Argument must be a 1-dimensional
    TypeError
        Elements of argument must be integers or floats

    Returns
    -------
    maximum : int or float
        Largest value from the data set

    See Also
    --------
    :func:`~regressions.statistics.sort.sorted_list`, :func:`~regressions.statistics.minimum.minimum_value`, :func:`~regressions.statistics.median.median_value`

    Notes
    -----
    - Set of numbers: :math:`a_i = \\{ a_1, a_2, \\cdots, a_n \\}`
    - Maximum value of set: :math:`a_{max} \\geq a_j, \\forall a_j \\in a_i`
    - |maximum|

    Examples
    --------
    Import `maximum_value` function from `regressions` library
        >>> from regressions.statistics.maximum import maximum_value
    Determine the maximum of the set [21, 53, 3, 68, 43, 9, 72, 19, 20, 1]
        >>> maximum_even = maximum_value([21, 53, 3, 68, 43, 9, 72, 19, 20, 1])
        >>> print(maximum_even)
        72
    Determine the maximum of the set [12, 81, 13, 8, 42, 72, 91, 20, 20]
        >>> maximum_odd = maximum_value([12, 81, 13, 8, 42, 72, 91, 20, 20])
        >>> print(maximum_odd)
        91
    """
    # Handle input errors
    vector_of_scalars(data)

    # Sort input
    sorted_data = sorted_list(data)

    # Grab last element of sorted input
    result = sorted_data[-1]
    return result


from regressions.errors.vectors import vector_of_scalars

def sum_value(data):
    """
    Calculates the sum of all elements in a data set

    Parameters
    ----------
    data : list of int or float
        List of numbers

    Raises
    ------
    TypeError
        Argument must be a 1-dimensional list
    TypeError
        Elements of argument must be integers or floats

    Returns
    -------
    total : float
        Number representing the sum of all elements in the original set

    See Also
    --------
    :func:`~regressions.statistics.mean.mean_value`, :func:`~regressions.statistics.correlation.correlation_coefficient`

    Notes
    -----
    - Set of numbers: :math:`a_i = \\{ a_1, a_2, \\cdots, a_n \\}`
    - Sum of all numbers in set: :math:`\\sum\\limits_{i=1}^n a_i = a_1 + a_2 + \\cdots + a_n`
    - |summation_notation|

    Examples
    --------
    Import `sum_value` function from `regressions` library
        >>> from regressions.statistics.summation import sum_value
    Find the total sum of all values in the array [2, 3, 5, 7]
        >>> total_1 = sum_value([2, 3, 5, 7])
        >>> print(total_1)
        17.0
    Find the total sum of all values in the array [1, -1, 1, -1]
        >>> total_2 = sum_value([1, -1, 1, -1])
        >>> print(total_2)
        0.0
    """
    # Handle input errors
    vector_of_scalars(data)

    # Create number to return
    result = 0

    # Iterate over input
    for element in data:
        # Add each element from input to number to return
        result += element
    
    # Convert number to float
    floated_result = float(result)
    return floated_result


from regressions.errors.analyses import select_equations
from regressions.errors.vectors import vector_of_scalars
from regressions.errors.scalars import select_integers, positive_integer
from regressions.statistics.sort import sorted_list
from regressions.vectors.separate import separate_elements
from .derivatives.linear import linear_derivatives
from .derivatives.quadratic import quadratic_derivatives
from .derivatives.cubic import cubic_derivatives
from .derivatives.hyperbolic import hyperbolic_derivatives
from .derivatives.exponential import exponential_derivatives
from .derivatives.logarithmic import logarithmic_derivatives
from .derivatives.logistic import logistic_derivatives
from .derivatives.sinusoidal import sinusoidal_derivatives
from .criticals import critical_points

def sign_chart(equation_type, coefficients, derivative_level, precision = 4):
    """
    Creates a sign chart for a given derivative

    Parameters
    ----------
    equation_type : str
        Name of the type of function for which the sign chart must be constructed (e.g., 'linear', 'quadratic')
    coefficients : list of int or float
        Coefficients to use to generate the equation to investigate
    derivative_level : int
        Integer corresponding to which derivative to investigate for sign chart (1 for the first derivative and 2 for the second derivative)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the results

    Raises
    ------
    ValueError
        First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'
    TypeError
        Second argument must be a 1-dimensional list containing elements that are integers or floats
    ValueError
        Third argument must be one of the following integers: [1, 2]
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    chart : list of str and float
        Strings describing the sign (e.g., 'positive', 'negative') of the derivative between its critical points; as a result, its elements will alternate between strings (indicating the signs) and floats (indicating the end points); if the function is sinusoidal, then only the initial results within a two-period interval will be listed, but a general form to determine other end points will also be included

    See Also
    --------
    - Roots for key functions: :func:`~regressions.analyses.roots.linear.linear_roots`, :func:`~regressions.analyses.roots.quadratic.quadratic_roots`, :func:`~regressions.analyses.roots.cubic.cubic_roots`, :func:`~regressions.analyses.roots.hyperbolic.hyperbolic_roots`, :func:`~regressions.analyses.roots.exponential.exponential_roots`, :func:`~regressions.analyses.roots.logarithmic.logarithmic_roots`, :func:`~regressions.analyses.roots.logistic.logistic_roots`, :func:`~regressions.analyses.roots.sinusoidal.sinusoidal_roots`
    - Graphical analysis: :func:`~regressions.analyses.criticals.critical_points`, :func:`~regressions.analyses.points.key_coordinates`

    Notes
    -----
    - Critical points for the derivative of a function: :math:`c_i = \\{ c_1, c_2, c_3,  \\cdots, c_{n-1}, c_n \\}`
    - Midpoints and key values of the intervals demarcated by the critical points: :math:`m_i = \\{ c_1 - 1,  \\frac{c_1+c_2}{2}, \\frac{c_2+c_3}{2}, \\cdots, \\frac{c_{n-1}+c_n}{2}, c_n + 1 \\}`
    - Values of the derivative within the intervals: :math:`v_i = \\{ v_1, v_2, v_3, v_4 \\cdots, v_{n-1}, v_n, v_{n+1} \\}`
    - Sign chart: :math:`s = ( v_1, c_1, v_2, c_2, v_3, c_3, v_4, \\dots, v_{n-1}, c_{n-1}, v_n, c_n, v_{n+1} )`

        - :math:`v_j = negative` if :math:`f'(m_j) < 0`
        - :math:`v_j = constant` if :math:`f'(m_j) = 0`
        - :math:`v_j = positve` if :math:`f'(m_j) > 0`

    - |intervals|

    Examples
    --------
    Import `sign_chart` function from `regressions` library
        >>> from regressions.analyses.intervals import sign_chart
    Create the sign chart for the first derivative of a cubic function with coefficients 1, -15, 63, and -7
        >>> chart_cubic = sign_chart('cubic', [1, -15, 63, -7], 1)
        >>> print(chart_cubic)
        ['positive', 3.0, 'negative', 7.0, 'positive']
    Create the sign chart for the second derivative of a sinusoidal function with coefficients 2, 3, 5, and 7
        >>> chart_sinusoidal = sign_chart('sinusoidal', [2, 3, 5, 7], 2)
        >>> print(chart_sinusoidal)
        ['positive', 5.0, 'negative', 6.0472, 'positive', 7.0944, 'negative', 8.1416, 'positive', 9.1888, 'negative', '5.0 + 1.0472k']
    """
    # Handle input errors
    select_equations(equation_type)
    vector_of_scalars(coefficients, 'second')
    select_integers(derivative_level, [1, 2], 'third')
    positive_integer(precision)

    # Create first and second derivatives based on equation type
    both_derivatives = {}
    if equation_type == 'linear':
        both_derivatives = linear_derivatives(*coefficients, precision)
    elif equation_type == 'quadratic':
        both_derivatives = quadratic_derivatives(*coefficients, precision)
    elif equation_type == 'cubic':
        both_derivatives = cubic_derivatives(*coefficients, precision)
    elif equation_type == 'hyperbolic':
        both_derivatives = hyperbolic_derivatives(*coefficients, precision)
    elif equation_type == 'exponential':
        both_derivatives = exponential_derivatives(*coefficients, precision)
    elif equation_type == 'logarithmic':
        both_derivatives = logarithmic_derivatives(*coefficients, precision)
    elif equation_type == 'logistic':
        both_derivatives = logistic_derivatives(*coefficients, precision)
    elif equation_type == 'sinusoidal':
        both_derivatives = sinusoidal_derivatives(*coefficients, precision)
    
    # Grab specific derivative evaluation based on derivative level
    derivative = lambda x : x
    if derivative_level == 1:
        derivative = both_derivatives['first']['evaluation']
    elif derivative_level == 2:
        derivative = both_derivatives['second']['evaluation']
    
    # Create critical points for specific derivative
    points = critical_points(equation_type, coefficients, derivative_level, precision)
    
    # Create sign chart for specific derivative
    result = []

    # Handle no critical points
    if points[0] == None:
        # Test an arbitrary number
        if derivative(10) > 0:
            result = ['positive']
        elif derivative(10) < 0:
            result = ['negative']
        else:
            result = ['constant']
    
    # Handle exactly one critical point
    elif len(points) == 1:
        # Generate numbers to test
        turning_point = points[0]
        before = turning_point - 1
        after = turning_point + 1

        # Test numbers
        if derivative(before) > 0:
            before = 'positive'
        elif derivative(before) < 0:
            before = 'negative'
        if derivative(after) > 0 :
            after = 'positive'
        elif derivative(after) < 0:
            after = 'negative'
        
        # Store sign chart
        result = [before, turning_point, after]
    
    # Handle exactly two critical points
    elif len(points) == 2:
        # Generate numbers to test
        sorted_points = sorted_list(points)
        first_point = sorted_points[0]
        second_point = sorted_points[1]
        middle = (first_point + second_point) / 2
        before = first_point - 1
        after = second_point + 1

        # Test numbers
        if derivative(before) > 0:
            before = 'positive'
        elif derivative(before) < 0:
            before = 'negative'
        if derivative(middle) > 0 :
            middle = 'positive'
        elif derivative(middle) < 0:
            middle = 'negative'
        if derivative(after) > 0:
            after = 'positive'
        elif derivative(after) < 0:
            after = 'negative'
        
        # Store sign chart
        result = [before, first_point, middle, second_point, after]
    
    # Handle more than two critical points
    else:
        # Separate numerical inputs from string inputs
        separated_points = separate_elements(points)
        numerical_points = separated_points['numerical']
        other_points = separated_points['other']
        
        # Generate numbers to test
        sorted_points = sorted_list(numerical_points)
        difference = sorted_points[1] - sorted_points[0]
        halved_difference = difference / 2
        before_first = sorted_points[0] - halved_difference
        between_first_second = sorted_points[0] + halved_difference
        between_second_third = sorted_points[1] + halved_difference
        between_third_fourth = sorted_points[2] + halved_difference
        between_fourth_last = sorted_points[3] + halved_difference
        after_last = sorted_points[4] + halved_difference
        test_points = [before_first, between_first_second, between_second_third, between_third_fourth, between_fourth_last, after_last]

        # Test numbers
        signs = []
        for point in test_points:
            if derivative(point) > 0:
                signs.append('positive')
            elif derivative(point) < 0:
                signs.append('negative')
        
        # Store sign chart
        result = [signs[0], sorted_points[0], signs[1], sorted_points[1], signs[2], sorted_points[2], signs[3], sorted_points[3], signs[4], sorted_points[4], signs[5], *other_points]
    return result


from math import exp
from regressions.errors.scalars import three_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value, rounded_list

def logistic_derivatives(first_constant, second_constant, third_constant, precision = 4):
    """
    Calculates the first and second derivatives of a logistic function

    Parameters
    ----------
    first_constant : int or float
        Carrying capacity of the original logistic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Growth rate of the original logistic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    third_constant : int or float
        Value of the sigmoid's midpoint of the original logistic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First three arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    derivatives['first']['constants'] : list of float
        Coefficients of the resultant first derivative
    derivatives['first']['evaluation'] : func
        Function for evaluating the resultant first derivative at any float or integer argument
    derivatives['second']['constants'] : list of float
        Coefficients of the resultant second derivative
    derivatives['second']['evaluation'] : func
        Function for evaluating the resultant second derivative at any float or integer argument

    See Also
    --------
    :func:`~regressions.analyses.equations.logistic.logistic_equation`, :func:`~regressions.analyses.integrals.logistic.logistic_integral`, :func:`~regressions.analyses.roots.logistic.logistic_roots`, :func:`~regressions.models.logistic.logistic_model`

    Notes
    -----
    - Standard form of a logistic function: :math:`f(x) = \\frac{a}{1 + \\text{e}^{-b\\cdot(x - c)}}`
    - First derivative of a logistic function: :math:`f'(x) = \\frac{ab\\cdot{\\text{e}^{-b\\cdot(x - c)}}}{(1 + \\text{e}^{-b\\cdot(x - c)})^2}`
    - Second derivative of a logistic function: :math:`f''(x) = \\frac{2ab^2\\cdot{\\text{e}^{-2b\\cdot(x - c)}}}{(1 + \\text{e}^{-b\\cdot(x - c)})^3} - \\frac{ab^2\\cdot{\\text{e}^{-b\\cdot(x - c)}}}{(1 + \\text{e}^{-b\\cdot(x - c)})^2}`
    - |differentiation_formulas|
    - |chain_rule|
    - |exponential|

    Examples
    --------
    Import `logistic_derivatives` function from `regressions` library
        >>> from regressions.analyses.derivatives.logistic import logistic_derivatives
    Generate the derivatives of a logistic function with coefficients 2, 3, and 5, then display the coefficients of its first and second derivatives
        >>> derivatives_constants = logistic_derivatives(2, 3, 5)
        >>> print(derivatives_constants['first']['constants'])
        [6.0, 3.0, 5.0]
        >>> print(derivatives_constants['second']['constants'])
        [18.0, 3.0, 5.0]
    Generate the derivatives of a logistic function with coefficients 100, 5, and 11, then evaluate its first and second derivatives at 10
        >>> derivatives_evaluation = logistic_derivatives(100, 5, 11)
        >>> print(derivatives_evaluation['first']['evaluation'](10))
        3.324
        >>> print(derivatives_evaluation['second']['evaluation'](10))
        16.3977
    Generate the derivatives of a logistic function with all inputs set to 0, then display the coefficients of its first and second derivatives
        >>> derivatives_zeroes = logistic_derivatives(0, 0, 0)
        >>> print(derivatives_zeroes['first']['constants'])
        [0.0001, 0.0001, 0.0001]
        >>> print(derivatives_zeroes['second']['constants'])
        [0.0001, 0.0001, 0.0001]
    """
    # Handle input errors
    three_scalars(first_constant, second_constant, third_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant, third_constant], precision)

    # Create first derivative
    first_coefficients = [coefficients[0] * coefficients[1], coefficients[1], coefficients[2]]
    first_constants = rounded_list(first_coefficients, precision)
    def first_derivative(variable):
        exponential = exp(-1 * first_constants[1] * (variable - first_constants[2]))
        evaluation = first_constants[0] * exponential * (1 + exponential)**(-2)
        rounded_evaluation = rounded_value(evaluation, precision)
        return rounded_evaluation
    first_dictionary = {
        'constants': first_constants,
        'evaluation': first_derivative
    }

    # Create second derivative
    second_coefficients = [first_constants[0] * first_constants[1], first_constants[1], first_constants[2]]
    second_constants = rounded_list(second_coefficients, precision)
    def second_derivative(variable):
        exponential = exp(-1 * second_constants[1] * (variable - second_constants[2]))
        evaluation = second_constants[0] * exponential * (1 + exponential)**(-2) * (2 * exponential / (1 + exponential) - 1)
        rounded_evaluation = rounded_value(evaluation, precision)
        return rounded_evaluation
    second_dictionary = {
        'constants': second_constants,
        'evaluation': second_derivative
    }
    
    # Package both derivatives in single dictionary
    results = {
        'first': first_dictionary,
        'second': second_dictionary
    }
    return results


from regressions.errors.scalars import three_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value, rounded_list

def quadratic_integral(first_constant, second_constant, third_constant, precision = 4):
    """
    Generates the integral of a quadratic function

    Parameters
    ----------
    first_constant : int or float
        Coefficient of the quadratic term of the original quadratic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Coefficient of the linear term of the original quadratic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    third_constant : int or float
        Coefficient of the constant term of the original quadratic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First three arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    integral['constants'] : list of float
        Coefficients of the resultant integral
    integral['evaluation'] : func
        Function for evaluating the resultant integral at any float or integer argument

    See Also
    --------
    :func:`~regressions.analyses.equations.quadratic.quadratic_equation`, :func:`~regressions.analyses.derivatives.quadratic.quadratic_derivatives`, :func:`~regressions.analyses.roots.quadratic.quadratic_roots`, :func:`~regressions.models.quadratic.quadratic_model`

    Notes
    -----
    - Standard form of a quadratic function: :math:`f(x) = a\\cdot{x^2} + b\\cdot{x} + c`
    - Integral of a quadratic function: :math:`F(x) = \\frac{a}{3}\\cdot{x^3} + \\frac{b}{2}\\cdot{x^2} + c\\cdot{x}`
    - |indefinite_integral|
    - |integration_formulas|

    Examples
    --------
    Import `quadratic_integral` function from `regressions` library
        >>> from regressions.analyses.integrals.quadratic import quadratic_integral
    Generate the integral of a quadratic function with coefficients 2, 3, and 5, then display its coefficients
        >>> integral_constants = quadratic_integral(2, 3, 5)
        >>> print(integral_constants['constants'])
        [0.6667, 1.5, 5.0]
    Generate the integral of a quadratic function with coefficients 7, -5, and 3, then evaluate its integral at 10
        >>> integral_evaluation = quadratic_integral(7, -5, 3)
        >>> print(integral_evaluation['evaluation'](10))
        2113.3
    Generate the integral of a quadratic function with all inputs set to 0, then display its coefficients
        >>> integral_zeroes = quadratic_integral(0, 0, 0)
        >>> print(integral_zeroes['constants'])
        [0.0001, 0.0001, 0.0001]
    """
    # Handle input errors
    three_scalars(first_constant, second_constant, third_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant, third_constant], precision)

    # Create constants
    integral_coefficients = [(1/3) * coefficients[0], (1/2) * coefficients[1], coefficients[2]]
    constants = rounded_list(integral_coefficients, precision)

    # Create evaluation
    def quadratic_evaluation(variable):
        evaluation = constants[0] * variable**3 + constants[1] * variable**2 + constants[2] * variable
        rounded_evaluation = rounded_value(evaluation, precision)
        return rounded_evaluation
    
    # Package constants and evaluation in single dictionary
    results = {
        'constants': constants,
        'evaluation': quadratic_evaluation
    }
    return results


from math import atan, degrees
from regressions.errors.vectors import vector_of_scalars, length

def vector_direction(vector):
    """
    Calculates the direction of a vector in radians and degrees

    Parameters
    ----------
    vector : list of int or float
        List of two numbers representing a vector, in which the first number is the horizontal component and the second is the vertical component

    Raises
    ------
    TypeError
        Argument must be a 1-dimensional list
    TypeError
        Elements of argument must be integers or floats
    ValueError
        Argument must contain exactly two elements

    Returns
    -------
    direction['radian'] : float
        Measure of the angle of the vector in radians
    direction['degree'] : float
        Measure of the angle of the vector in degrees

    See Also
    --------
    :func:`~regressions.vectors.components.component_form`, :func:`~regressions.vectors.magnitude.vector_magnitude`, :func:`~regressions.vectors.unit.unit_vector`

    Notes
    -----
    - Vector: :math:`\\langle x, y \\rangle`
    - Direction of vector: :math:`\\theta = \\tan^{-1}(\\frac{y}{x})`
    - |direction|

    Examples
    --------
    Import `vector_direction` function from `regressions` library
        >>> from regressions.vectors.direction import vector_direction
    Determine the direction of a vector with a component form of [7, 5]
        >>> direction_positive = vector_direction([7, 5])
        >>> print(direction_positive['radian'])
        0.6202494859828215
        >>> print(direction_positive['degree'])
        35.53767779197438
    Determine the direction of a vector with a component form of [-3, 11]
        >>> direction_negative = vector_direction([-3, 11])
        >>> print(direction_negative['radian'])
        -1.3045442776439713
        >>> print(direction_negative['degree'])
        -74.74488129694222
    """
    # Handle input errors
    vector_of_scalars(vector)
    length(vector, 2)

    # Circumvent division by zero
    if vector[0] == 0:
        vector[0] = 0.0001
    
    # Create intermediary variable
    ratio = vector[1] / vector[0]
    
    # Determine direction in radians
    radian_measure = atan(ratio)

    # Convert radians into degrees
    degree_measure = degrees(radian_measure)
    
    # Package both measures in single dictionary
    result = {
        'radian': radian_measure,
        'degree': degree_measure
    }
    return result


from math import log
from regressions.errors.scalars import two_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value, rounded_list

def exponential_integral(first_constant, second_constant, precision = 4):
    """
    Generates the integral of an exponential function

    Parameters
    ----------
    first_constant : int or float
        Constant multiple of the original exponential function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Base rate of variable of the original exponential function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001); if one, it will be converted to a small, near-one decimal value (e.g., 1.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First two arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    integral['constants'] : list of float
        Coefficients of the resultant integral
    integral['evaluation'] : func
        Function for evaluating the resultant integral at any float or integer argument

    See Also
    --------
    :func:`~regressions.analyses.equations.exponential.exponential_equation`, :func:`~regressions.analyses.derivatives.exponential.exponential_derivatives`, :func:`~regressions.analyses.roots.exponential.exponential_roots`, :func:`~regressions.models.exponential.exponential_model`

    Notes
    -----
    - Standard form of an exponential function: :math:`f(x) = a\\cdot{b^x}`
    - Integral of an exponential function: :math:`F(x) = \\frac{a}{\\ln{b}}\\cdot{b^x}`
    - |indefinite_integral|
    - |integration_formulas|

    Examples
    --------
    Import `exponential_integral` function from `regressions` library
        >>> from regressions.analyses.integrals.exponential import exponential_integral
    Generate the integral of an exponential function with coefficients 2 and 3, then display its coefficients
        >>> integral_constants = exponential_integral(2, 3)
        >>> print(integral_constants['constants'])
        [1.8205, 3.0]
    Generate the integral of an exponential function with coefficients -2 and 3, then evaluate its integral at 10
        >>> integral_evaluation = exponential_integral(-2, 3)
        >>> print(integral_evaluation['evaluation'](10))
        -107498.7045
    Generate the integral of an exponential function with all inputs set to 0, then display its coefficients
        >>> integral_zeroes = exponential_integral(0, 0)
        >>> print(integral_zeroes['constants'])
        [-0.0001, 0.0001]
    """
    # Handle input errors
    two_scalars(first_constant, second_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant], precision)

    # Circumvent division by zero
    if coefficients[1] == 1:
        coefficients[1] = 1 + 10**(-precision)
    
    # Create constants
    integral_coefficients = [coefficients[0] / log(abs(coefficients[1])), coefficients[1]]
    constants = rounded_list(integral_coefficients, precision)

    # Create evaluation
    def exponential_evaluation(variable):
        evaluation = constants[0] * constants[1]**variable
        rounded_evaluation = rounded_value(evaluation, precision)
        return rounded_evaluation
    
    # Package constants and evaluation in single dictionary
    results = {
        'constants': constants,
        'evaluation': exponential_evaluation
    }
    return results


from regressions.errors.scalars import two_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value

def exponential_equation(first_constant, second_constant, precision = 4):
    """
    Generates an exponential function to provide evaluations at variable inputs

    Parameters
    ----------
    first_constant : int or float
        Constant multiple of the resultant exponential function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Base rate of variable of the resultant exponential function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First two arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    evaluation : func
        Function for evaluating an exponential equation when passed any integer or float argument
    
    See Also
    --------
    :func:`~regressions.analyses.derivatives.exponential.exponential_derivatives`, :func:`~regressions.analyses.integrals.exponential.exponential_integral`, :func:`~regressions.analyses.roots.exponential.exponential_roots`, :func:`~regressions.models.exponential.exponential_model`

    Notes
    -----
    - Standard form of an exponential function: :math:`f(x) = a\\cdot{b^x}`
    - |exponential_functions|

    Examples
    --------
    Import `exponential_equation` function from `regressions` library
        >>> from regressions.analyses.equations.exponential import exponential_equation
    Create an exponential function with coefficients 2 and 3, then evaluate it at 10
        >>> evaluation_first = exponential_equation(2, 3)
        >>> print(evaluation_first(10))
        118098.0
    Create an exponential function with coefficients -2 and 3, then evaluate it at 10
        >>> evaluation_second = exponential_equation(-2, 3)
        >>> print(evaluation_second(10))
        -118098.0
    Create an exponential function with all inputs set to 0, then evaluate it at 10
        >>> evaluation_zero = exponential_equation(0, 0)
        >>> print(evaluation_zero(10))
        0.0001
    """
    # Handle input errors
    two_scalars(first_constant, second_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant], precision)

    # Create evaluation
    def exponential_evaluation(variable):
        evaluation = coefficients[0] * coefficients[1]**variable
        result = rounded_value(evaluation, precision)
        return result
    return exponential_evaluation


from regressions.errors.vectors import allow_none_vector

def separate_elements(vector):
    """
    Separates a single vector into two vectors based their data types (numerical and non-numerical)

    Parameters
    ----------
    vector : list of int or float or str
        List containing elements of various types: integers, floats, strings, or None

    Raises
    ------
    TypeError
        Elements of argument must be integers, floats, strings, or None

    Returns
    -------
    separate_vectors['numerical'] : list of int or float
        List containing only the integer or float elements from the input vector; may be an empty list if input vector contained no integer or float elements
    separate_vectors['other'] : list of str
        List containing only the string or None elements from the input vector; may be an empty list if input vector contained no string elements and did not merely contain None

    See Also
    --------
    :func:`~regressions.vectors.generate_generate_elements`, :func:`~regressions.vectors.unify.unite_vectors`

    Notes
    -----
    - Set of all real numbers: :math:`\\mathbb{R}`
    - Set of all strings: :math:`\\mathbb{S}`
    - Set containing None: :math:`\\{ \\emptyset \\}`
    - Set of all real numbers, all strings, and None: :math:`\\mathbb{A} = \\{ \\mathbb{R}, \\mathbb{S}, \\emptyset \\}`
    - Set of mixed elements: :math:`M = \\{ m \\mid m \\in \\mathbb{A} \\}`
    - Set of numerical elements from mixed elements: :math:`N = \\{ n \\mid n \\in M, n \\in \\mathbb{R} \\}`
    - Set of non-numerical elements from mixed elements: :math:`O = \\{ o \\mid o \\in M, o \\in \\mathbb{S} \\cup \\emptyset \\}`

    Examples
    --------
    Import `separate_elements` function from `regressions` library
        >>> from regressions.vectors.separate import separate_elements
    Separate [1, 'two', 3, 'four'] into two vectors based on its data types
        >>> separate_vectors_mixed = separate_elements([1, 'two', 3, 'four'])
        >>> print(separate_vectors_mixed['numerical'])
        [1, 3]
        >>> print(separate_vectors_mixed['other'])
        ['two', 'four']
    Separate [None] into two vectors based on its data types
        >>> separate_vectors_none = separate_elements([None])
        >>> print(separate_vectors_none['numerical'])
        []
        >>> print(separate_vectors_none['other'])
        [None]
    """
    # Handle input errors
    allow_none_vector(vector)

    # Create lists to store separate values
    numerical_elements = []
    other_elements = []

    # Iterate over input
    for element in vector:
        # Store numerical elements
        if isinstance(element, (int, float)):
            numerical_elements.append(element)
        
        # Store non-numerical elements
        else:
            other_elements.append(element)
    
    # Package both types of element in single dictionary to return
    results = {
        'numerical': numerical_elements,
        'other': other_elements
    }
    return results


from regressions.errors.scalars import four_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value, rounded_list

def cubic_integral(first_constant, second_constant, third_constant, fourth_constant, precision = 4):
    """
    Generates the integral of a cubic function

    Parameters
    ----------
    first_constant : int or float
        Coefficient of the cubic term of the original cubic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Coefficient of the quadratic term of the original cubic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    third_constant : int or float
        Coefficient of the linear term of the original cubic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    fourth_constant : int or float
        Coefficient of the constant term of the original cubic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First four arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    integral['constants'] : list of float
        Coefficients of the resultant integral
    integral['evaluation'] : func
        Function for evaluating the resultant integral at any float or integer argument
    
    See Also
    --------
    :func:`~regressions.analyses.equations.cubic.cubic_equation`, :func:`~regressions.analyses.derivatives.cubic.cubic_derivatives`, :func:`~regressions.analyses.roots.cubic.cubic_roots`, :func:`~regressions.models.cubic.cubic_model`

    Notes
    -----
    - Standard form of a cubic function: :math:`f(x) = a\\cdot{x^3} + b\\cdot{x^2} + c\\cdot{x} + d`
    - Integral of a cubic function: :math:`F(x) = \\frac{a}{4}\\cdot{x^4} + \\frac{b}{3}\\cdot{x^3} + \\frac{c}{2}\\cdot{x^2} + d\\cdot{x}`
    - |indefinite_integral|
    - |integration_formulas|

    Examples
    --------
    Import `cubic_integral` function from `regressions` library
        >>> from regressions.analyses.integrals.cubic import cubic_integral
    Generate the integral of a cubic function with coefficients 2, 3, 5, and 7, then display its coefficients
        >>> integral_constants = cubic_integral(2, 3, 5, 7)
        >>> print(integral_constants['constants'])
        [0.5, 1.0, 2.5, 7.0]
    Generate the integral of a cubic function with coefficients 7, -5, -3, and 2, then evaluate its integral at 10
        >>> integral_evaluation = cubic_integral(7, -5, -3, 2)
        >>> print(integral_evaluation['evaluation'](10))
        15703.3
    Generate the integral of a cubic function with all inputs set to 0, then display its coefficients
        >>> integral_zeroes = cubic_integral(0, 0, 0, 0)
        >>> print(integral_zeroes['constants'])
        [0.0001, 0.0001, 0.0001, 0.0001]
    """
    # Handle input errors
    four_scalars(first_constant, second_constant, third_constant, fourth_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant, third_constant, fourth_constant], precision)

    # Generate constants
    integral_coefficients = [(1/4) * coefficients[0], (1/3) * coefficients[1], (1/2) * coefficients[2], coefficients[3]]
    constants = rounded_list(integral_coefficients, precision)

    # Create evaluation
    def cubic_evaluation(variable):
        evaluation = constants[0] * variable**4 + constants[1] * variable**3 + constants[2] * variable**2 + constants[3] * variable
        rounded_evaluation = rounded_value(evaluation, precision)
        return rounded_evaluation
    
    # Package constants and evaluation in single dictionary
    results = {
        'constants': constants,
        'evaluation': cubic_evaluation
    }
    return results


from regressions.errors.vectors import vector_of_scalars
from regressions.statistics.summation import sum_value

def vector_magnitude(vector):
    """
    Calculates the magnitude of a vector

    Parameters
    ----------
    vector : list of int or float
        List of numbers representing a vector

    Raises
    ------
    TypeError
        Argument must be a 1-dimensional list
    TypeError
        Elements of argument must be integers or floats

    Returns
    -------
    magnitude : float
        Measure of the size of the vector, as determined by taking the root of the sum of the squares of its components

    See Also
    --------
    :func:`~regressions.vectors.components.component_form`, :func:`~regressions.vectors.direction.vector_direction`,
    :func:`~regressions.vectors.unit.unit_vector`

    Notes
    -----
    - Vector: :math:`\\mathbf{a} = \\langle a_1, a_2, \\cdots, a_n \\rangle`
    - Magnitude of vector: :math:`\\|\\mathbf{a}\\| = \\sqrt{a_1^2 + a_2^2 + \\cdots + a_n^2}`
    - |magnitude|

    Examples
    --------
    Import `vector_magnitude` function from `regressions` library
        >>> from regressions.vectors.magnitude import vector_magnitude
    Determine the magnitude of the vector with components [7, 5, -1]
        >>> magnitude_3d = vector_magnitude([7, 5, -1])
        >>> print(magnitude_3d)
        8.660254037844387
    Determine the magnitude of the vector with components [3, 2]
        >>> magnitude_2d = vector_magnitude([3, 2])
        >>> print(magnitude_2d)
        3.605551275463989
    """
    # Handle input errors
    vector_of_scalars(vector)

    # Create intermediary list
    squares = []

    # Iterate over input
    for element in vector:
        # Store squares of each element in intermediary list
        squares.append(element**2)
    
    # Add all squares in list
    sum_squares = sum_value(squares)

    # Take the square root of the sum of all squares
    result = sum_squares**(1/2)
    return result


from math import log
from regressions.errors.scalars import three_scalars, four_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.sort import sorted_list
from regressions.statistics.rounding import rounded_value
from .quadratic import quadratic_roots

def logistic_roots(first_constant, second_constant, third_constant, precision = 4):
    """
    Calculates the roots of a logistic function

    Parameters
    ----------
    first_constant : int or float
        Carrying capacity of the original logistic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Growth rate of the original logistic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    third_constant : int or float
        Value of the sigmoid's midpoint of the original logistic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First three arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    roots : list of float
        List of the x-coordinates of all of the x-intercepts of the original function; if the function never crosses the x-axis, then it will return a list of `None`

    See Also
    --------
    :func:`~regressions.analyses.equations.logistic.logistic_equation`, :func:`~regressions.analyses.derivatives.logistic.logistic_derivatives`, :func:`~regressions.analyses.integrals.logistic.logistic_integral`, :func:`~regressions.models.logistic.logistic_model`

    Notes
    -----
    - Standard form of a logistic function: :math:`f(x) = \\frac{a}{1 + \\text{e}^{-b\\cdot(x - c)}}`
    - Logistic formula: :math:`x = \\varnothing`

    Examples
    --------
    Import `logistic_roots` function from `regressions` library
        >>> from regressions.analyses.roots.logistic import logistic_roots
    Calculate the roots of a logistic function with coefficients 2, 3, and 5
        >>> roots_first = logistic_roots(2, 3, 5)
        >>> print(roots_first)
        [None]
    Calculate the roots of a logistic function with coefficients 100, 5, and 11
        >>> roots_second = logistic_roots(100, 5, 11)
        >>> print(roots_second)
        [None]
    Calculate the roots of a logistic function with all inputs set to 0
        >>> roots_zeroes = logistic_roots(0, 0, 0)
        >>> print(roots_zeroes)
        [None]
    """
    # Handle input errors
    three_scalars(first_constant, second_constant, third_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant, third_constant], precision)

    # Create list to return
    result = []

    # Determine root
    root = None

    # Return result
    result.append(root)
    return result

def logistic_roots_first_derivative(first_constant, second_constant, third_constant, precision = 4):
    # Handle input errors
    three_scalars(first_constant, second_constant, third_constant)
    positive_integer(precision)
    
    # Create list to return
    result = []

    # Determine root of first derivative
    root = None

    # Return result
    result.append(root)
    return result

def logistic_roots_second_derivative(first_constant, second_constant, third_constant, precision = 4):
    # Handle input errors
    three_scalars(first_constant, second_constant, third_constant)
    positive_integer(precision)

    # Create list to return
    result = []

    # Determine root of second derivative
    root = rounded_value(third_constant)

    # Return root
    result.append(root)
    return result

def logistic_roots_initial_value(first_constant, second_constant, third_constant, initial_value, precision = 4):
    # Handle input errors
    four_scalars(first_constant, second_constant, third_constant, initial_value)
    positive_integer(precision)

    # Create list to return
    result = []
    
    # Create pivot variable
    log_argument = first_constant / initial_value - 1

    # Circumvent logarithm of zero
    if log_argument == 0:
        log_argument = 10**(-precision)

    # Create intermediary variables
    numerator = log(abs(log_argument))
    denominator = second_constant
    ratio = numerator / denominator

    # Determine root given an initial value
    root = third_constant - ratio

    # Round root
    rounded_root = rounded_value(root, precision)

    # Return result
    result.append(rounded_root)
    return result

def logistic_roots_derivative_initial_value(first_constant, second_constant, third_constant, initial_value, precision = 4):
    # Handle input errors
    four_scalars(first_constant, second_constant, third_constant, initial_value)
    positive_integer(precision)

    # Create intermediary list and list to return
    roots = []
    result = []

    # Determine quadratic roots of derivative given an initial value
    intermediary_roots = quadratic_roots(initial_value, 2 * initial_value - first_constant * second_constant, initial_value, precision)

    # Handle no roots
    if intermediary_roots[0] == None:
        roots.append(None)
    
    # Convert quadratic roots using logarithms
    else:
        for intermediary in intermediary_roots:
            root = third_constant - log(abs(intermediary)) / second_constant
            rounded_root = rounded_value(root, precision)
            roots.append(rounded_root)
    
    # Sort roots
    sorted_roots = sorted_list(roots)

    # Return result
    result.extend(sorted_roots)
    return result


from regressions.errors.vectors import vector_of_scalars
from .magnitude import vector_magnitude
from .multiplication import scalar_product_vector

def unit_vector(vector):
    """
    Calculates the unit vector corresponding to a given vector (and therefore having the same direction)

    Parameters
    ----------
    vector : list of int or float
        List of numbers representing a vector

    Raises
    ------
    TypeError
        Argument must be a 1-dimensional list
    TypeError
        Elements of argument must be integers or floats

    Returns
    -------
    unit : list of float
        Vector with a magnitue of 1 in the same direction as the original vector

    See Also
    --------
    :func:`~regressions.vectors.components.component_form`, :func:`~regressions.vectors.multiplication.scalar_product_vector`, :func:`~regressions.vectors.direction.vector_direction`,
    :func:`~regressions.vectors.magnitude.vector_magnitude`

    Notes
    -----
    - Comparison vector: :math:`\\mathbf{a} = \\langle a_1, a_2, \\cdots, a_n \\rangle`
    - Unit vector with same direction: :math:`\\mathbf{u}= \\frac{\\mathbf{a}}{\\|\\mathbf{a}\\|}`
    - |unit_vector|

    Examples
    --------
    Import `unit_vector` function from `regressions` library
        >>> from regressions.vectors.unit import unit_vector
    Determine the unit vector of the vector with components [7, 5, -1]
        >>> unit_3d = unit_vector([7, 5, -1])
        >>> print(unit_3d)
        [0.8082903768654759, 0.5773502691896257, -0.11547005383792514]
    Determine the unit vector of the vector with components [3, 2]
        >>> unit_2d = unit_vector([3, 2])
        >>> print(unit_2d)
        [0.8320502943378437, 0.5547001962252291]
    """
    # Handle input errors
    vector_of_scalars(vector)

    # Determine magnitude of input
    magnitude = vector_magnitude(vector)

    # Circumvent division by zero
    if magnitude == 0:
        magnitude = 0.0001
    
    # Calculate reciprocal of magnitude
    reciprocal_magnitude = 1 / magnitude

    # Multiply input by the reciprocal of its magnitude
    result = scalar_product_vector(vector, reciprocal_magnitude)
    return result


from regressions.errors.scalars import scalar_value
from regressions.errors.vectors import vector_of_scalars, compare_vectors

def scalar_product_vector(vector, scalar):
    """
    Calculates the product of a vector and a scalar

    Parameters
    ----------
    vector : list of int or float
        List of numbers representing a vector
    scalar : int or float
        Number representing a scalar

    Raises
    ------
    TypeError
        First argument must be a 1-dimensional list
    TypeError
        Elements of first argument must be integers or floats
    TypeError
        Second argument must be an integer or a float

    Returns
    -------
    product : list of int or float
        List of numbers in which each element is the product of the scalar factor and the corresponding element from the input vector

    See Also
    --------
    :func:`~regressions.matrices.multiplication.scalar_product_matrix`, :func:`~regressions.vectors.addition.vector_sum`

    Notes
    -----
    - Vector: :math:`\\mathbf{a} = \\langle a_1, a_2, \\cdots, a_n \\rangle`
    - Scalar: :math:`c`
    - Scalar product: :math:`c\\cdot{\\mathbf{a}} = \\langle c\\cdot{a_1}, c\\cdot{a_2}, \\cdots, c\\cdot{a_n} \\rangle`
    - |scalar_multiplication|

    Examples
    --------
    Import `scalar_product_vector` function from `regressions` library
        >>> from regressions.vectors.multiplication import scalar_product_vector
    Multiply [1, 2, 3] and -2
        >>> product_3d = scalar_product_vector([1, 2, 3], -2)
        >>> print(product_3d)
        [-2, -4, -6]
    Multiply [-5, 12] and 3
        >>> product_2d = scalar_product_vector([-5, 12], 3)
        >>> print(product_2d)
        [-15, 36]
    """
    # Handle input errors
    vector_of_scalars(vector, 'first')
    scalar_value(scalar, 'second')

    # Create list to return
    result = []

    # Iterate over input
    for element in vector:
        # Store products in list to return
        result.append(element * scalar)
    
    # Return result
    return result

def dot_product(vector_one, vector_two):
    """
    Calculates the product of two vectors

    Parameters
    ----------
    vector_one : list of int or float
        List of numbers representing a vector
    vector_two : list of int or float
        List of numbers representing a vector

    Raises
    ------
    TypeError
        Arguments must be 1-dimensional lists
    TypeError
        Elements of arguments must be integers or floats
    ValueError
        Both arguments must contain the same number of elements

    Returns
    -------
    product : float
        Number created by summing the products of the corresponding terms from each input vector

    See Also
    --------
    :func:`~regressions.matrices.multiplication.matrix_product`

    Notes
    -----
    - First vector: :math:`\\mathbf{a} = \\langle a_1, a_2, \\cdots, a_n \\rangle`
    - Second vector: :math:`\\mathbf{b} = \\langle b_1, b_2, \\cdots, b_n \\rangle`:
    - Dot product of vectors: :math:`\\mathbf{a}\\cdot{\\mathbf{b}} = a_1\\cdot{b_1} + a_2\\cdot{b_2} + \\cdots + a_n\\cdot{b_n}`
    - |dot_product|

    Examples
    --------
    Import `dot_product` function from `regressions` library
        >>> from regressions.vectors.multiplication import dot_product
    Multiply [1, 2, 3] and [4, 5, 6]
        >>> product_3d = dot_product([1, 2, 3], [4, 5, 6])
        >>> print(product_3d)
        32.0
    Multiply [-5, 12] and [3, -7]
        >>> product_2d = dot_product([-5, 12], [3, -7])
        >>> print(product_2d)
        -99.0
    """
    # Handle input errors
    compare_vectors(vector_one, vector_two)

    # Create intermediary number
    result = 0

    # Iterate over inputs
    for i in range(len(vector_one)):
        # Add products of corresponding elements from inputs to intermediary number
        result += vector_one[i] * vector_two[i]
    
    # Convert number to float
    floated_result = float(result)
    return floated_result


from regressions.errors.vectors import vector_of_scalars
from regressions.errors.scalars import select_integers
from .median import median_value
from .halve import half

def quartile_value(data, q):
    """
    Determines the first, second, or third quartile values of a data set

    Parameters
    ----------
    data : list of int or float
        List of numbers to analyze
    q : int
        Number determining which quartile to provide

    Raises
    ------
    TypeError
        First argument must be a 1-dimensional list
    TypeError
        Elements of first argument must be integers or floats
    ValueError
        Second argument must be an integer contained within the set [1, 2, 3]

    Returns
    -------
    quartile : int or float
        Quartile value of the data set

    See Also
    --------
    :func:`~regressions.statistics.sort.sorted_list`, :func:`~regressions.statistics.halve.half`, :func:`~regressions.statistics.minimum.minimum_value`, :func:`~regressions.statistics.maximum.maximum_value`, :func:`~regressions.statistics.median.median_value`

    Notes
    -----
    - Ordered set of numbers: :math:`a_i = ( a_1, a_2, \\cdots, a_n )`
    - For sets with an odd amount of numbers:
        
        - First quartile: :math:`Q_1 = a_{\\lceil n/4 \\rceil}`
        - Second quartile: :math:`Q_2 = a_{\\lceil n/2 \\rceil}`
        - Third quartile: :math:`Q_3 = a_{\\lceil 3n/4 \\rceil}`
    
    - For sets with an even amount of numbers:

        - If :math:`n \\text{ mod } 4 \\neq 0`:

            - First quartile: :math:`Q_1 = a_{\\lceil n/4 \\rceil}`
            - Second quartile: :math:`Q_2 = \\frac{a_{n/2} + a_{n/2 + 1}}{2}`
            - Third quartile: :math:`Q_3 = a_{\\lceil 3n/4 \\rceil}`
        
        - If :math:`n \\text{ mod } 4 = 0`:

            - First quartile: :math:`Q_1 = \\frac{a_{n/4} + a_{n/4 + 1}}{2}`
            - Second quartile: :math:`Q_2 = \\frac{a_{n/2} + a_{n/2 + 1}}{2}`
            - Third quartile: :math:`Q_3 = \\frac{a_{3n/4} + a_{3n/4 + 1}}{2}`

    - |quartiles|

    Examples
    --------
    Import `quartile_value` function from `regressions` library
        >>> from regressions.statistics.quartiles import quartile_value
    Determine the first quartile of the set [21, 53, 3, 68, 43, 9, 72, 19, 20, 1]
        >>> quartile_1 = quartile_value([21, 53, 3, 68, 43, 9, 72, 19, 20, 1], 1)
        >>> print(quartile_1)
        9
    Determine the third quartile of the set [12, 81, 13, 8, 42, 72, 91, 20, 20]
        >>> quartile_3 = quartile_value([12, 81, 13, 8, 42, 72, 91, 20, 20], 3)
        >>> print(quartile_3)
        76.5
    """
    # Handle input errors
    vector_of_scalars(data, 'first')
    select_integers(q, [1, 2, 3])

    # Split input in half
    halved_data = half(data)

    # Create number to return
    result = 0

    # Determine Q2 by taking the median of all elements
    if q == 2:
        result = median_value(data)
    
    # Determine Q1 by taking the median of the lower half of elements
    elif q == 1:
        result = median_value(halved_data['lower'])
    
    # Determine Q3 by taking the median of the upper half of elements
    elif q == 3:
        result = median_value(halved_data['upper'])
    
    # Return result
    return result


from math import exp, log
from regressions.errors.scalars import three_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value, rounded_list

def logistic_integral(first_constant, second_constant, third_constant, precision = 4):
    """
    Generates the integral of a logistic function

    Parameters
    ----------
    first_constant : int or float
        Carrying capacity of the original logistic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Growth rate of the original logistic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    third_constant : int or float
        Value of the sigmoid's midpoint of the original logistic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First three arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    integral['constants'] : list of float
        Coefficients of the resultant integral
    integral['evaluation'] : func
        Function for evaluating the resultant integral at any float or integer argument

    See Also
    --------
    :func:`~regressions.analyses.equations.logistic.logistic_equation`, :func:`~regressions.analyses.derivatives.logistic.logistic_derivatives`, :func:`~regressions.analyses.roots.logistic.logistic_roots`, :func:`~regressions.models.logistic.logistic_model`

    Notes
    -----
    - Standard form of a logistic function: :math:`f(x) = \\frac{a}{1 + \\text{e}^{-b\\cdot(x - c)}}`
    - Integral of a logistic function: :math:`F(x) = \\frac{a}{b}\\cdot{\\ln|\\text{e}^{b\\cdot(x - c)} + 1|}`
    - |indefinite_integral|
    - |integration_formulas|
    - |substitution_rule|

    Examples
    --------
    Import `logistic_integral` function from `regressions` library
        >>> from regressions.analyses.integrals.logistic import logistic_integral
    Generate the integral of a logistic function with coefficients 2, 3, and 5, then display its coefficients
        >>> integral_constants = logistic_integral(2, 3, 5)
        >>> print(integral_constants['constants'])
        [0.6667, 3.0, 5.0]
    Generate the integral of a logistic function with coefficients 100, 5, and 11, then evaluate its integral at 10
        >>> integral_evaluation = logistic_integral(100, 5, 11)
        >>> print(integral_evaluation['evaluation'](10))
        0.1343
    Generate the integral of a logistic function with all inputs set to 0, then display its coefficients
        >>> integral_zeroes = logistic_integral(0, 0, 0)
        >>> print(integral_zeroes['constants'])
        [1.0, 0.0001, 0.0001]
    """
    # Handle input errors
    three_scalars(first_constant, second_constant, third_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant, third_constant], precision)

    # Create constants
    integral_coefficients = [coefficients[0] / coefficients[1], coefficients[1], coefficients[2]]
    constants = rounded_list(integral_coefficients, precision)

    # Create evaluation
    def logistic_evaluation(variable):
        evaluation = constants[0] * log(abs(exp(constants[1] * (variable - constants[2])) + 1))
        rounded_evaluation = rounded_value(evaluation, precision)
        return rounded_evaluation
    
    # Package constants and evaluation in single dictionary
    results = {
        'constants': constants,
        'evaluation': logistic_evaluation
    }
    return results


from regressions.errors.matrices import matrix_of_scalars, level
from regressions.errors.scalars import positive_integer

def single_dimension(matrix, scalar = 1):
    """
    Extracts a column vector as a row vector from a matrix according to an integer corresponding to the column's position

    Parameters
    ----------
    matrix : list of lists of int or float
        List containing other lists, where each inner list is a row and elements within those inner lists correspond to columns
    scalar : int, default=1
        Number corresponding to the column's position

    Raises
    ------
    TypeError
        First argument must be a 2-dimensional list
    TypeError
        Elements nested within the first argument's lists must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    vector : list of int or float
        List containing only integers or floats

    See Also
    --------
    :func:`~regressions.vectors.column.column_conversion`, :func:`~regressions.statistics.sort.sorted_dimension`, :func:`~regressions.statistics.half.half_dimension`

    Notes
    -----
    - Matrix: :math:`\\begin{bmatrix} a_{1,1} & a_{1,2} & \\cdots & a_{1,n} \\\\ a_{2,1} & a_{2,2} & \\cdots & a_{2,n} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ a_{m,1} & a_{m,2} & \\cdots & a_{m,n} \\end{bmatrix}`
    - Row vector corresponding to the :math:`n`\ th column of the matrix: :math:`\\langle a_{1,n}, a_{2,n}, \\cdots, a_{m,n} \\rangle`

    Examples
    --------
    Import `single_dimension` function from `regressions` library
        >>> from regressions.vectors.dimension import single_dimension
    Extract the second column from the matrix [[3, 5, 9], [1, -4, 2]]
        >>> vector_2c = single_dimension([[3, 5, 9], [1, -4, 2]], 2)
        >>> print(vector_2c)
        [5, -4]
    Extract the first column from the matrix [[3, 5, 9], [1, -4, 2]]
        >>> vector_1c = single_dimension([[3, 5, 9], [1, -4, 2]], 1)
        >>> print(vector_1c)
        [3, 1]
    """
    # Handle input errors
    matrix_of_scalars(matrix, 'first')
    positive_integer(scalar)
    level(matrix, scalar)

    # Create list to return
    result = []

    # Iterate over input
    for element in matrix:
        # Store all elements at given dimension in list to return
        result.append(element[scalar - 1])
    
    # Return result
    return result


from math import sin
from regressions.errors.scalars import four_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value

def sinusoidal_equation(first_constant, second_constant, third_constant, fourth_constant, precision = 4):
    """
    Generates a sinusoidal function to provide evaluations at variable inputs

    Parameters
    ----------
    first_constant : int or float
        Vertical stretch factor of the resultant sine function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Horizontal stretch factor of the resultant sine function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    third_constant : int or float
        Horizontal shift of the resultant sine function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    fourth_constant : int or float
        Vertical shift of the resultant sine function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First four arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    evaluation : func
        Function for evaluating a sinusoidal equation when passed any integer or float argument

    See Also
    --------
    :func:`~regressions.analyses.derivatives.sinusoidal.sinusoidal_derivatives`, :func:`~regressions.analyses.integrals.sinusoidal.sinusoidal_integral`, :func:`~regressions.analyses.roots.sinusoidal.sinusoidal_roots`, :func:`~regressions.models.sinusoidal.sinusoidal_model`

    Notes
    -----
    - Standard form of a sinusoidal function: :math:`f(x) = a\\cdot{\\sin(b\\cdot(x - c))} + d`

        - Period of function: :math:`\\frac{2\\pi}{|b|}`
        - Amplitude of function: :math:`|a|`

    - |sine_functions|

    Examples
    --------
    Import `sinusoidal_equation` function from `regressions` library
        >>> from regressions.analyses.equations.sinusoidal import sinusoidal_equation
    Create a sinusoidal function with coefficients 2, 3, 5, and 7, then evaluate it at 10
        >>> evaluation_first = sinusoidal_equation(2, 3, 5, 7)
        >>> print(evaluation_first(10))
        8.3006
    Create a sinusoidal function with coefficients 7, -5, -3, and 2, then evaluate it at 10
        >>> evaluation_second = sinusoidal_equation(7, -5, -3, 2)
        >>> print(evaluation_second(10))
        -3.7878
    Create a sinusoidal function with all inputs set to 0, then evaluate it at 10
        >>> evaluation_zero = sinusoidal_equation(0, 0, 0, 0)
        >>> print(evaluation_zero(10))
        0.0001
    """
    # Handle input errors
    four_scalars(first_constant, second_constant, third_constant, fourth_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant, third_constant, fourth_constant], precision)

    # Create evaluation
    def sinusoidal_evaluation(variable):
        evaluation = coefficients[0] * sin(coefficients[1] * (variable - coefficients[2])) + coefficients[3]
        result = rounded_value(evaluation, precision)
        return result
    return sinusoidal_evaluation


from regressions.errors.analyses import select_equations
from regressions.errors.scalars import compare_scalars, positive_integer
from regressions.errors.vectors import vector_of_scalars
from regressions.statistics.rounding import rounded_value
from .roots.linear import linear_roots_initial_value, linear_roots_derivative_initial_value
from .roots.quadratic import quadratic_roots_initial_value, quadratic_roots_derivative_initial_value
from .roots.cubic import cubic_roots_initial_value, cubic_roots_derivative_initial_value
from .roots.hyperbolic import hyperbolic_roots_initial_value, hyperbolic_roots_derivative_initial_value
from .roots.exponential import exponential_roots_initial_value, exponential_roots_derivative_initial_value
from .roots.logarithmic import logarithmic_roots_initial_value, logarithmic_roots_derivative_initial_value
from .roots.logistic import logistic_roots_initial_value, logistic_roots_derivative_initial_value
from .roots.sinusoidal import sinusoidal_roots_initial_value, sinusoidal_roots_derivative_initial_value
from .equations.linear import linear_equation
from .equations.quadratic import quadratic_equation
from .equations.cubic import cubic_equation
from .equations.hyperbolic import hyperbolic_equation
from .equations.exponential import exponential_equation
from .equations.logarithmic import logarithmic_equation
from .equations.logistic import logistic_equation
from .equations.sinusoidal import sinusoidal_equation
from .accumulation import accumulated_area
from .points import points_within_range, shifted_points_within_range

def average_value_derivative(equation_type, coefficients, start, end, precision = 4):
    """
    Evaluates the average rate of change between two points for a given function

    Parameters
    ----------
    equation_type : str
        Name of the type of function for which the definite integral must be evaluated (e.g., 'linear', 'quadratic')
    coefficients : list of int or float
        Coefficients of the original function to use for evaluating the average rate of change
    start : int or float
        Value of the x-coordinate of the first point to use for evaluating the rate of change
    end : int or float
        Value of the x-coordinate of the second point to use for evaluating the rate of change
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the result

    Raises
    ------
    ValueError
        First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'
    TypeError
        Second argument must be a 1-dimensional list containing elements that are integers or floats
    TypeError
        Third and fourth arguments must be integers or floats
    ValueError
        Third argument must be less than or equal to fourth argument
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    average : float
        Slope of a function between two points; if start and end values are identical, then slope will be zero

    See Also
    --------
    :func:`~regressions.analyses.mean_values.mean_values_derivative`, 
    :func:`~regressions.analyses.mean_values.average_value_integral`, :func:`~regressions.analyses.mean_values.mean_values_integral`

    Notes
    -----
    - Slope of a function over an interval: :math:`m = \\frac{f(b) - f(a)}{b - a}`
    - |mean_derivatives|

    Examples
    --------
    Import `average_value_derivative` function from `regressions` library
        >>> from regressions.analyses.mean_values import average_value_derivative
    Evaluate the average rate of change of a cubic function with coefficients 2, 3, 5, and 7 between end points of 10 and 20
        >>> average_cubic = average_value_derivative('cubic', [2, 3, 5, 7], 10, 20)
        >>> print(average_cubic)
        1495.0
    Evaluate the average rate of change of a sinusoidal function with coefficients 2, 3, 5, and 7 between end points of 10 and 20
        >>> average_sinusoidal = average_value_derivative('sinusoidal', [2, 3, 5, 7], 10, 20)
        >>> print(average_sinusoidal)
        0.0401
    """
    # Handle input errors
    select_equations(equation_type)
    vector_of_scalars(coefficients, 'second')
    compare_scalars(start, end, 'third', 'fourth')
    positive_integer(precision)

    # Create equation based on equation type
    equation = lambda x : x
    if equation_type == 'linear':
        equation = linear_equation(*coefficients, precision)
    elif equation_type == 'quadratic':
        equation = quadratic_equation(*coefficients, precision)
    elif equation_type == 'cubic':
        equation = cubic_equation(*coefficients, precision)
    elif equation_type == 'hyperbolic':
        equation = hyperbolic_equation(*coefficients, precision)
    elif equation_type == 'exponential':
        equation = exponential_equation(*coefficients, precision)
    elif equation_type == 'logarithmic':
        equation = logarithmic_equation(*coefficients, precision)
    elif equation_type == 'logistic':
        equation = logistic_equation(*coefficients, precision)
    elif equation_type == 'sinusoidal':
        equation = sinusoidal_equation(*coefficients, precision)
    
    # Create intermediary variables
    vertical_change = equation(end) - equation(start)
    horizontal_change = end - start
    
    # Circumvent division by zero
    if horizontal_change == 0:
        horizontal_change = 10**(-precision)
    
    # Determine average slope
    ratio = vertical_change / horizontal_change
    result = rounded_value(ratio, precision)
    return result

def mean_values_derivative(equation_type, coefficients, start, end, precision = 4):
    """
    Generates a list of all the x-coordinates whose instantaneous rates of change equal the function's average rate of change between two points

    Parameters
    ----------
    equation_type : str
        Name of the type of function for which an average value must be determined (e.g., 'linear', 'quadratic')
    coefficients : list of int or float
        Coefficients to use to generate the equation to investigate
    start : int or float
        Value of the x-coordinate of the first point to use for evaluating the rate of change; all results must be greater than this value
    end : int or float
        Value of the x-coordinate of the second point to use for evaluating the rate of change; all results must be less than this value
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the result

    Raises
    ------
    ValueError
        First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'
    TypeError
        Second argument must be a 1-dimensional list containing elements that are integers or floats
    TypeError
        Third and fourth arguments must be integers or floats
    ValueError
        Third argument must be less than or equal to fourth argument
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    points : list of float or str
        Values of the x-coordinates within the specified interval at which the original function has an instantaneous rate of change equal to its average rate of change over that entire interval; if the function is sinusoidal, then only the initial results within at most a four-period interval within the specified interval will be listed, but general forms will also be included (however, their results may be outside the specified interval); if the algorithm cannot determine any values, then it will return a list of `None`

    See Also
    --------
    - Roots for key functions: :func:`~regressions.analyses.roots.linear.linear_roots`, :func:`~regressions.analyses.roots.quadratic.quadratic_roots`, :func:`~regressions.analyses.roots.cubic.cubic_roots`, :func:`~regressions.analyses.roots.hyperbolic.hyperbolic_roots`, :func:`~regressions.analyses.roots.exponential.exponential_roots`, :func:`~regressions.analyses.roots.logarithmic.logarithmic_roots`, :func:`~regressions.analyses.roots.logistic.logistic_roots`, :func:`~regressions.analyses.roots.sinusoidal.sinusoidal_roots`
    - Mean values: :func:`~regressions.analyses.mean_values.average_value_derivative`, :func:`~regressions.analyses.mean_values.average_value_integral`, :func:`~regressions.analyses.mean_values.mean_values_integral`

    Notes
    -----
    - Mean values for the derivative over an interval: :math:`f'(c) = \\frac{f(b) - f(a)}{b - a}` 
    - |mean_derivatives|

    Examples
    --------
    Import `mean_values_derivative` function from `regressions` library
        >>> from regressions.analyses.mean_values import mean_values_derivative
    Generate a list of all the x-coordinates whose instantaneous rates of change equal the function's average rate of change for a cubic function with coefficients 2, 3, 5, and 7 between end points of 10 and 20
        >>> points_cubic = mean_values_derivative('cubic', [2, 3, 5, 7], 10, 20)
        >>> print(points_cubic)
        [15.2665]
    Generate a list of all the x-coordinates whose instantaneous rates of change equal the function's average rate of change for a cubic function with coefficients 2, 3, 5, and 7 between end points of 10 and 20
        >>> points_sinusoidal = mean_values_derivative('sinusoidal', [2, 3, 5, 7], 10, 20)
        >>> print(points_sinusoidal)
        [10.7618, 11.8046, 12.8562, 13.899, 14.9506, 15.9934, 17.045, 18.0878, 19.1394, '10.7618 + 2.0944k', '11.8046 + 2.0944k']
    """
    # Handle input errors
    select_equations(equation_type)
    vector_of_scalars(coefficients, 'second')
    compare_scalars(start, end, 'third', 'fourth')
    positive_integer(precision)

    # Create list to return
    result = []

    # Determine average rate of change over interval
    average = average_value_derivative(equation_type, coefficients, start, end, precision)

    # Circumvent division by zero
    if average == 0:
        average = 10**(-precision)

    # Determine points satisfying theorem based on equation type
    if equation_type == 'linear':
        result = linear_roots_derivative_initial_value(*coefficients, average, precision)
    elif equation_type == 'quadratic':
        result = quadratic_roots_derivative_initial_value(*coefficients, average, precision)
    elif equation_type == 'cubic':
        result = cubic_roots_derivative_initial_value(*coefficients, average, precision)
    elif equation_type == 'hyperbolic':
        result = hyperbolic_roots_derivative_initial_value(*coefficients, average, precision)
    elif equation_type == 'exponential':
        result = exponential_roots_derivative_initial_value(*coefficients, average, precision)
    elif equation_type == 'logarithmic':
        result = logarithmic_roots_derivative_initial_value(*coefficients, average, precision)
    elif equation_type == 'logistic':
        result = logistic_roots_derivative_initial_value(*coefficients, average, precision)
    elif equation_type == 'sinusoidal':
        options = sinusoidal_roots_derivative_initial_value(*coefficients, average, precision)
        result = shifted_points_within_range(options, start, end, precision)
    
    # Eliminate points outside of interval
    final_result = points_within_range(result, start, end)
    return final_result

def average_value_integral(equation_type, coefficients, start, end, precision = 4):
    """
    Evaluates the average value of a given function between two points

    Parameters
    ----------
    equation_type : str
        Name of the type of function for which the definite integral must be evaluated (e.g., 'linear', 'quadratic')
    coefficients : list of int or float
        Coefficients of the original function to integrate
    start : int or float
        Value of the x-coordinate of the first point to use for evaluating the average value
    end : int or float
        Value of the x-coordinate of the second point to use for evaluating the average value
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the result

    Raises
    ------
    ValueError
        First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'
    TypeError
        Second argument must be a 1-dimensional list containing elements that are integers or floats
    TypeError
        Third and fourth arguments must be integers or floats
    ValueError
        Third argument must be less than or equal to fourth argument
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    average : float
        Average value of the function between two points; if start and end values are identical, then average value will be zero

    See Also
    --------
    :func:`~regressions.analyses.accumulation.accumulated_area`, :func:`~regressions.analyses.mean_values.mean_values_integral`, :func:`~regressions.analyses.mean_values.average_value_derivative`, :func:`~regressions.analyses.mean_values.mean_values_derivative`

    Notes
    -----
    - Average value of a function over an interval: :math:`f_{avg} = \\frac{1}{b - a}\\cdot{\\int_{a}^{b} f(x) \\,dx}`
    - |mean_integrals|

    Examples
    --------
    Import `average_value_integral` function from `regressions` library
        >>> from regressions.analyses.mean_values import average_value_integral
    Evaluate the average value of a cubic function with coefficients 2, 3, 5, and 7 between end points of 10 and 20
        >>> average_cubic = average_value_integral('cubic', [2, 3, 5, 7], 10, 20)
        >>> print(average_cubic)
        8282.0
    Evaluate the average value of a sinusoidal function with coefficients 2, 3, 5, and 7 between end points of 10 and 20
        >>> average_sinusoidal = average_value_integral('sinusoidal', [2, 3, 5, 7], 10, 20)
        >>> print(average_sinusoidal)
        6.9143
    """
    # Handle input errors
    select_equations(equation_type)
    vector_of_scalars(coefficients, 'second')
    compare_scalars(start, end, 'third', 'fourth')
    positive_integer(precision)

    # Determine accumulated value of function over interval
    accumulated_value = accumulated_area(equation_type, coefficients, start, end, precision)

    # Create intermediary variable
    change = end - start

    # Circumvent division by zero
    if change == 0:
        change = 10**(-precision)
    
    # Determine average value of function over interval
    ratio = accumulated_value / change

    # Round value
    rounded_ratio = rounded_value(ratio, precision)

    # Return result
    result = rounded_ratio
    return result

def mean_values_integral(equation_type, coefficients, start, end, precision = 4):
    """
    Generates a list of all the x-coordinates between two points at which a function's value will equal its average value over that interval

    Parameters
    ----------
    equation_type : str
        Name of the type of function for which an average value must be determined (e.g., 'linear', 'quadratic')
    coefficients : list of int or float
        Coefficients of the origianl function under investigation
    start : int or float
        Value of the x-coordinate of the first point to use for evaluating the average value; all results must be greater than this value
    end : int or float
        Value of the x-coordinate of the second point to use for evaluating the average value; all results must be less than this value
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the result

    Raises
    ------
    ValueError
        First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'
    TypeError
        Second argument must be a 1-dimensional list containing elements that are integers or floats
    TypeError
        Third and fourth arguments must be integers or floats
    ValueError
        Third argument must be less than or equal to fourth argument
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    points : list of float or str
        Values of the x-coordinates within the specified interval at which the original function has a value equal to its average value over that entire interval; if the function is sinusoidal, then only the initial results within at most a four-period interval within the specified interval will be listed, but general forms will also be included (however, their results may be outside the specified interval); if the algorithm cannot determine any values, then it will return a list of `None`

    See Also
    --------
    - Roots for key functions: :func:`~regressions.analyses.roots.linear.linear_roots`, :func:`~regressions.analyses.roots.quadratic.quadratic_roots`, :func:`~regressions.analyses.roots.cubic.cubic_roots`, :func:`~regressions.analyses.roots.hyperbolic.hyperbolic_roots`, :func:`~regressions.analyses.roots.exponential.exponential_roots`, :func:`~regressions.analyses.roots.logarithmic.logarithmic_roots`, :func:`~regressions.analyses.roots.logistic.logistic_roots`, :func:`~regressions.analyses.roots.sinusoidal.sinusoidal_roots`
    - Mean values: :func:`~regressions.analyses.mean_values.average_value_integral`, :func:`~regressions.analyses.mean_values.average_value_derivative`, :func:`~regressions.analyses.mean_values.mean_values_derivative`

    Notes
    -----
    - Mean values for the integral over an interval: :math:`f(c) = \\frac{1}{b - a}\\cdot{\\int_{a}^{b} f(x) \\,dx}` 
    - |mean_integrals|

    Examples
    --------
    Import `mean_values_integral` function from `regressions` library
        >>> from regressions.analyses.mean_values import mean_values_integral
    Generate a list of all the x-coordinates of a cubic function with coefficients 2, 3, 5, and 7 at which the function's value will equal its average value between 10 and 20
        >>> points_cubic = mean_values_integral('cubic', [2, 3, 5, 7], 10, 20)
        >>> print(points_cubic)
        [15.5188]
    Generate a list of all the x-coordinates of a sinusoidal function with coefficients 2, 3, 5, and 7 at which the function's value will equal its average value between 10 and 20
        >>> points_sinusoidal = mean_values_integral('sinusoidal', [2, 3, 5, 7], 10, 20)
        >>> print(points_sinusoidal)
        [10.2503, 11.2689, 12.3447, 13.3633, 14.4391, 15.4577, 16.5335, 17.5521, 18.6279, 19.6465, '10.2503 + 2.0944k', '11.2689 + 2.0944k']
    """
    # Handle input errors
    select_equations(equation_type)
    vector_of_scalars(coefficients, 'second')
    compare_scalars(start, end, 'third', 'fourth')
    positive_integer(precision)

    # Create list to return
    result = []

    # Determine average value of function over interval
    average = average_value_integral(equation_type, coefficients, start, end, precision)

    # Circumvent division by zero
    if average == 0:
        average = 10**(-precision)
    
    # Determine points satisfying theorem based on equation type
    if equation_type == 'linear':
        result = linear_roots_initial_value(*coefficients, average, precision)
    elif equation_type == 'quadratic':
        result = quadratic_roots_initial_value(*coefficients, average, precision)
    elif equation_type == 'cubic':
        result = cubic_roots_initial_value(*coefficients, average, precision)
    elif equation_type == 'hyperbolic':
        result = hyperbolic_roots_initial_value(*coefficients, average, precision)
    elif equation_type == 'exponential':
        result = exponential_roots_initial_value(*coefficients, average, precision)
    elif equation_type == 'logarithmic':
        result = logarithmic_roots_initial_value(*coefficients, average, precision)
    elif equation_type == 'logistic':
        result = logistic_roots_initial_value(*coefficients, average, precision)
    elif equation_type == 'sinusoidal':
        options = sinusoidal_roots_initial_value(*coefficients, average, precision)
        result = shifted_points_within_range(options, start, end, precision)
    
    # Eliminate points outside of interval
    final_result = points_within_range(result, start, end)
    return final_result

def average_values(equation_type, coefficients, start, end, precision = 4):
    """
    Calculates the average values for a specific function

    Parameters
    ----------
    equation_type : str
        Name of the type of function for which average values must be determined (e.g., 'linear', 'quadratic')
    coefficients : list of int or float
        Coefficients of the origianl function under investigation
    start : int or float
        Value of the x-coordinate of the first point to use for evaluating the average values; results within lists must be greater than this value
    end : int or float
        Value of the x-coordinate of the second point to use for evaluating the average values; results within lists must be less than this value
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the result

    Raises
    ------
    ValueError
        First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'
    TypeError
        Second argument must be a 1-dimensional list containing elements that are integers or floats
    TypeError
        Third and fourth arguments must be integers or floats
    ValueError
        Third argument must be less than or equal to fourth argument
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    averages['average_value_derivative'] : float
        Slope of a function between two points
    averages['mean_values_derivative'] : list of float or str
        Values of the x-coordinates within the specified interval at which the original function has a value equal to its average value over that entire interval; if the function is sinusoidal, then only the initial results within at most a four-period interval within the specified interval will be listed, but general forms will also be included (however, their results may be outside the specified interval); if the algorithm cannot determine any values, then it will return a list of `None`
    averages['average_value_integral'] : float
        Average value of the function between two points
    averages['mean_values_integral'] : list of float or str
        Values of the x-coordinates within the specified interval at which the original function has a value equal to its average value over that entire interval; if the function is sinusoidal, then only the initial results within at most a four-period interval within the specified interval will be listed, but general forms will also be included (however, their results may be outside the specified interval); if the algorithm cannot determine any values, then it will return a list of `None`

    See Also
    --------
    :func:`~regressions.analyses.mean_values.average_value_derivative`, :func:`~regressions.analyses.mean_values.mean_values_derivative`, :func:`~regressions.analyses.mean_values.average_value_integral`, :func:`~regressions.analyses.mean_values.mean_values_integral`

    Notes
    -----
    - |mean_derivatives|
    - |mean_integrals|

    Examples
    --------
    Import `average_values` function from `regressions` library
        >>> from regressions.analyses.mean_values import average_values
    Calculate the averages of a cubic function with coefficients 2, 3, 5, and 7 between 10 and 20
        >>> averages_cubic = average_values('cubic', [2, 3, 5, 7], 10, 20)
        >>> print(averages_cubic['average_value_derivative'])
        1495.0
        >>> print(averages_cubic['mean_values_derivative'])
        [15.2665]
        >>> print(averages_cubic['average_value_integral'])
        8282.0
        >>> print(averages_cubic['mean_values_integral'])
        [15.5188]
    Calculate the averages of a sinusoidal function with coefficients 2, 3, 5, and 7 between 10 and 20
        >>> averages_sinusoidal = average_values('sinusoidal', [2, 3, 5, 7], 10, 20)
        >>> print(averages_sinusoidal['average_value_derivative'])
        0.0401
        >>> print(averages_sinusoidal['mean_values_derivative'])
        [10.7618, 11.8046, 12.8562, 13.899, 14.9506, 15.9934, 17.045, 18.0878, 19.1394, '10.7618 + 2.0944k', '11.8046 + 2.0944k']
        >>> print(averages_sinusoidal['average_value_integral'])
        6.9143
        >>> print(averages_sinusoidal['mean_values_integral'])
        [10.2503, 11.2689, 12.3447, 13.3633, 14.4391, 15.4577, 16.5335, 17.5521, 18.6279, 19.6465, '10.2503 + 2.0944k', '11.2689 + 2.0944k']
    """
    # Handle input errors
    select_equations(equation_type)
    vector_of_scalars(coefficients, 'second')
    compare_scalars(start, end, 'third', 'fourth')
    positive_integer(precision)

    # Determine various mean values
    derivative_value = average_value_derivative(equation_type, coefficients, start, end, precision)
    derivative_inputs = mean_values_derivative(equation_type, coefficients, start, end, precision)
    integral_value = average_value_integral(equation_type, coefficients, start, end, precision)
    integral_inputs = mean_values_integral(equation_type, coefficients, start, end, precision)

    # Package all values in single dictionary
    results = {
        'average_value_derivative': derivative_value,
        'mean_values_derivative': derivative_inputs,
        'average_value_integral': integral_value,
        'mean_values_integral': integral_inputs
    }
    return results


from math import log
from regressions.errors.scalars import two_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value

def logarithmic_equation(first_constant, second_constant, precision = 4):
    """
    Generates a logarithmic function to provide evaluations at variable inputs

    Parameters
    ----------
    first_constant : int or float
        Coefficient of the logarithmic term of the resultant logarithmic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Coefficient of the constant term of the resultant logarithmic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First two arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    evaluation : func
        Function for evaluating a logarithmic equation when passed any integer or float argument; if zero inputted as argument, it will be converted to a small, non-zero decimal value (e.g., 0.0001)

    See Also
    --------
    :func:`~regressions.analyses.derivatives.logarithmic.logarithmic_derivatives`, :func:`~regressions.analyses.integrals.logarithmic.logarithmic_integral`, :func:`~regressions.analyses.roots.logarithmic.logarithmic_roots`, :func:`~regressions.models.logarithmic.logarithmic_model`

    Notes
    -----
    - Standard form of a logarithmic function: :math:`f(x) = a\\cdot{\\ln{x}} + b`
    - |logarithmic_functions|

    Examples
    --------
    Import `logarithmic_equation` function from `regressions` library
        >>> from regressions.analyses.equations.logarithmic import logarithmic_equation
    Create a logarithmic function with coefficients 2 and 3, then evaluate it at 10
        >>> evaluation_first = logarithmic_equation(2, 3)
        >>> print(evaluation_first(10))
        7.6052
    Create a logarithmic function with coefficients -2 and 3, then evaluate it at 10
        >>> evaluation_second = logarithmic_equation(-2, 3)
        >>> print(evaluation_second(10))
        -1.6052
    Create a logarithmic function with all inputs set to 0, then evaluate it at 10
        >>> evaluation_zero = logarithmic_equation(0, 0)
        >>> print(evaluation_zero(10))
        0.0003
    """
    # Handle input errors
    two_scalars(first_constant, second_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant], precision)

    # Create evaluation
    def logarithmic_evaluation(variable):
        # Circumvent logarithm of zero
        if variable == 0:
            variable = 10**(-precision)
        evaluation = coefficients[0] * log(abs(variable)) + coefficients[1]
        result = rounded_value(evaluation, precision)
        return result
    return logarithmic_evaluation


from .positions import argument_position

def select_equations(string):
    choices = ['linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', 'sinusoidal']
    if string not in choices:
        raise ValueError("First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'")
    else:
        return "First argument is either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'"

def select_points(string = 'point', position = 'only'):
    identifier = argument_position(position)
    choices = ['point', 'intercepts', 'maxima', 'minima', 'inflections']
    if string not in choices:
        raise ValueError(f"{identifier.capitalize()} must be either 'point', 'intercepts', 'maxima', 'minima', or 'inflections'")
    else:
        return f"{identifier.capitalize()} is either 'point', 'intercepts', 'maxima', 'minima', or 'inflections'"


from regressions.errors.analyses import select_equations
from regressions.errors.vectors import vector_of_scalars
from regressions.errors.scalars import positive_integer
from .intervals import sign_chart

def maxima_points(equation_type, coefficients, precision = 4):
    """
    Calculates the maxima of a specific function

    Parameters
    ----------
    equation_type : str
        Name of the type of function for which the maxima must be determined (e.g., 'linear', 'quadratic')
    coefficients : list of int or float
        Coefficients to use to generate the equation to investigate
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the results

    Raises
    ------
    ValueError
        First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'
    TypeError
        Second argument must be a 1-dimensional list containing elements that are integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    points : list of float
        Values of the x-coordinates at which the original function has a relative maximum; if the function is sinusoidal, then only two or three results within a two-period interval will be listed; if the function has no maxima, then it will return a list of `None`

    See Also
    --------
    - Roots for key functions: :func:`~regressions.analyses.roots.linear.linear_roots`, :func:`~regressions.analyses.roots.quadratic.quadratic_roots`, :func:`~regressions.analyses.roots.cubic.cubic_roots`, :func:`~regressions.analyses.roots.hyperbolic.hyperbolic_roots`, :func:`~regressions.analyses.roots.exponential.exponential_roots`, :func:`~regressions.analyses.roots.logarithmic.logarithmic_roots`, :func:`~regressions.analyses.roots.logistic.logistic_roots`, :func:`~regressions.analyses.roots.sinusoidal.sinusoidal_roots`
    - Graphical analysis: :func:`~regressions.analyses.criticals.critical_points`, :func:`~regressions.analyses.intervals.sign_chart`, :func:`~regressions.analyses.minima.minima_points`, :func:`~regressions.analyses.extrema.extrema_points`, :func:`~regressions.analyses.points.key_coordinates`

    Notes
    -----
    - Critical points for the derivative of a function: :math:`c_i = \\{ c_1, c_2, c_3,  \\cdots, c_{n-1}, c_n \\}`
    - X-coordinates of the maxima of the function: :math:`x_{max} = \\{ x \\mid x \\in c_i, f'(\\frac{c_{j-1} + c_j}{2}) > 0, f'(\\frac{c_j + c_{j+1}}{2}) < 0 \\}`
    - |maxima_values|

    Examples
    --------
    Import `maxima_points` function from `regressions` library
        >>> from regressions.analyses.maxima import maxima_points
    Calculate the maxima of a cubic function with coefficients 1, -15, 63, and -7
        >>> points_cubic = maxima_points('cubic', [1, -15, 63, -7])
        >>> print(points_cubic)
        [3.0]
    Calculate the maxima of a sinusoidal function with coefficients 2, 3, 5, and 7
        >>> points_sinusoidal = maxima_points('sinusoidal', [2, 3, 5, 7])
        >>> print(points_sinusoidal)
        [5.5236, 7.618, 9.7124]
    """
    # Handle input errors
    select_equations(equation_type)
    vector_of_scalars(coefficients, 'second')
    positive_integer(precision)
    
    # Create sign chart
    intervals = sign_chart(equation_type, coefficients, 1, precision)
    
    # Determine maxima
    result = []
    for i in range(len(intervals)):
        try:
            if intervals[i] == 'positive' and intervals[i + 2] == 'negative':
                result.append(intervals[i + 1])
        except IndexError:
            pass
    
    # Handle no maxima
    if len(result) == 0:
        result.append(None)
    return result


from regressions.errors.matrices import compare_rows
from regressions.errors.scalars import positive_integer
from regressions.statistics.rounding import rounded_list
from regressions.vectors.dimension import single_dimension
from .multiplication import matrix_product
from .transpose import transposed_matrix
from .inverse import inverse_matrix

def system_solution(matrix_one, matrix_two, precision = 4):
    """
    Solves a system of equations using matrices (independent matrix multiplied by resultant matrix equals dependent matrix)

    Parameters
    ----------
    matrix_one : list of lists of int or float
        List of lists of numbers representing the independent matrix of a system of equations
    matrix_two : list of lists of int or float
        List of lists of numbers representing the dependent matrix of a system of equations
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the results

    Raises
    ------
    TypeError
        First and second arguments must be 2-dimensional lists
    TypeError
        Elements nested within first and second arguments must be integers or floats
    ValueError
        First and second arguments must contain the same amount of lists
    ValueError
        Last argument must be a positive integer
    
    Returns
    -------
    solution : list of float
        Row vector of coefficients that if expressed as a column vector would satisfy the equation

    See Also
    --------
    :func:`~regressions.matrices.multiplication.matrix_product`, :func:`~regressions.matrices.transpose.transposed_matrix`, :func:`~regressions.matrices.determinant.linear_determinant`, :func:`~regressions.matrices.inverse.inverse_matrix`

    Notes
    -----
    - Independent matrix: :math:`\\mathbf{A} = \\begin{bmatrix} a_{1,1} & a_{1,2} & \\cdots & a_{1,n} \\\\ a_{2,1} & a_{2,2} & \\cdots & a_{2,n} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ a_{m,1} & a_{m,2} & \\cdots & a_{m,n} \\end{bmatrix}`
    - Dependent matrix: :math:`\\mathbf{B} = \\begin{bmatrix} b_{1,1} \\\\ b_{2,1} \\\\ \\cdots \\\\ b_{m,1} \\end{bmatrix}`
    - Variable matrix: :math:`\\mathbf{X} = \\begin{bmatrix} x_{1,1} \\\\ x_{2,1} \\\\ \\cdots \\\\ x_{m,1} \\end{bmatrix}`
    - System of equations in terms of matrices: :math:`\\mathbf{A}\\cdot{\\mathbf{X}} = \\mathbf{B}`
    - Solution of system of equations: :math:`\\mathbf{X} = \\mathbf{A}^{-1}\\cdot{\\mathbf{B}}`
    - |solve|

    Examples
    --------
    Import `system_solution` function from `regressions` library
        >>> from regressions.matrices.solve import system_solution
    Solve the system that has an independent matrix of [[2, 3], [1, -1]] and a dependent matrix of [[5], [1]]
        >>> solution_2values = system_solution([[2, 3], [1, -1]], [[5], [1]])
        >>> print(solution_2values)
        [1.6, 0.6]
    Solve the system that has an independent matrix of [[1, -2, 3], [-4, 5, -6], [7, -8, 9], [-10, 11, 12]] and a dependent matrix of [[2], [-3], [5], [-7]]
        >>> solution_3values = system_solution([[1, -2, 3], [-4, 5, -6], [7, -8, 9], [-10, 11, 12]], [[2], [-3], [5], [-7]])
        >>> print(solution_3values)
        [-0.8611, -1.3889, -0.0278]
    """
    # Handle input errors
    compare_rows(matrix_one, matrix_two)
    positive_integer(precision)

    # Multiply inverse of first matrix by second matrix
    transposition = transposed_matrix(matrix_one)
    product = matrix_product(transposition, matrix_one)
    inversion = inverse_matrix(product)
    second_product = matrix_product(inversion, transposition)
    solution_column = matrix_product(second_product, matrix_two)

    # Convert solution from a column vector into a row vector
    solution = single_dimension(solution_column, 1)

    # Round result
    result = rounded_list(solution, precision)
    return result


from regressions.errors.vectors import compare_vectors
from .addition import vector_sum
from .multiplication import scalar_product_vector

def component_form(initial_point, terminal_point):
    """
    Calculates the component form for a vector by using two points

    Parameters
    ----------
    initial_point : list of int or float
        List of numbers representing a point
    terminal_point : list of int or float
        List of numbers representing a point

    Raises
    ------
    TypeError
        Arguments must be 1-dimensional lists
    TypeError
        Elements of arguments must be integers or floats
    ValueError
        Both arguments must contain the same number of elements

    Returns
    -------
    components : list of int or float
        List in which each element is the difference of the corresponding elements from the input points (specifically, the change from the initial point to the terminal point)

    See Also
    --------
    :func:`~regressions.vectors.addition.vector_sum`, :func:`~regressions.vectors.multiplication.scalar_product_vector`, :func:`~regressions.vectors.direction.vector_direction`, :func:`~regressions.vectors.magnitude.vector_magnitude`

    Notes
    -----
    - Initial point: :math:`A = (a_1, a_2, \\cdots, a_n)`
    - Terminal point: :math:`B = (b_1, b_2, \\cdots, b_n)`
    - Component form of vector: :math:`\\overrightarrow{AB} = \\langle b_1 - a_1, b_2 - a_2, \\cdots, b_n - a_n \\rangle`
    - |component_form|

    Examples
    --------
    Import `component_form` function from `regressions` library
        >>> from regressions.vectors.components import component_form
    Determine the component form of a vector with an initial point of [1, 2, 3] and a terminal point of [4, 5, 6]
        >>> components_3d = component_form([1, 2, 3], [4, 5, 6])
        >>> print(components_3d)
        [3, 3, 3]
    Determine the component form of a vector with an initial point of [-5, 12] and a terminal point of [3, -7]
        >>> components_2d = component_form([-5, 12], [3, -7])
        >>> print(components_2d)
        [8, -19]
    """
    # Handle input errors
    compare_vectors(initial_point, terminal_point)

    # Determine difference between terminal point and initial point
    result = vector_sum(terminal_point, scalar_product_vector(initial_point, -1))
    return result


from regressions.errors.vectors import compare_vectors

def unite_vectors(vector_one, vector_two):
    """
    Unites two row vectors into a single matrix whose columns coincide with the input vectors

    Parameters
    ----------
    vector_one : list of int or float
        List of numbers representing a vector
    vector_two : list of int or float
        List of numbers representing a vector

    Raises
    ------
    TypeError
        Arguments must be 1-dimensional lists
    TypeError
        Elements of arguments must be integers or floats
    ValueError
        Both arguments must contain the same number of elements

    Returns
    -------
    matrix : list of lists of int or float
        List containing lists; length of outer list will equal lengths of supplied vectors; length of inner lists will equal two

    See Also
    --------
    :func:`~regressions.vectors.dimension.single_dimension`, :func:`~regressions.statistics.halve.partition`

    Notes
    -----
    - First vector: :math:`\\langle a_1, a_2, \\cdots, a_n \\rangle`
    - Second vector: :math:`\\langle b_1, b_2, \\cdots, b_n \\rangle`
    - Matrix unifying first and second vectors: :math:`\\begin{bmatrix} a_1 & b_1 \\\\ a_2 & b_2 \\\\ \\cdots & \\cdots \\\\ a_n & b_n \\end{bmatrix}`

    Examples
    --------
    Import `unite_vectors` function from `regressions` library
        >>> from regressions.vectors.unify import unite_vectors
    Unite [1, 2, 3] and [4, 5, 6]
        >>> matrix_3x2 = unite_vectors([1, 2, 3], [4, 5, 6])
        >>> print(matrix_2x3)
        [[1, 4], [2, 5], [3, 6]]
    Unite [-5, 12] and [3, -7]
        >>> matrix_2x2 = unite_vectors([-5, 12], [3, -7])
        >>> print(matrix_2x2)
        [[-5, 3], [12, -7]]
    """
    # Handle input errors
    compare_vectors(vector_one, vector_two)

    # Create list to return
    result = []

    # Handle no solution
    if vector_one[0] == None:
        result.append(None)
    
    # Handle general case
    else:
        # Iterate over inputs
        for i in range(len(vector_one)):
            # Store corresponding elements from inputs as lists within list to return
            result.append([vector_one[i], vector_two[i]])
    
    # Return result
    return result


from regressions.errors.scalars import scalar_value
from regressions.errors.vectors import compare_vectors

def single_residual(actual, expected):
    """
    Calculates the difference between the actual value and the expected value

    Parameters
    ----------
    actual : int or float
        Value actually provided by an initial data set
    expected : int or float
        Value predicted to occur by a generated model at the same input to the one that coincided with the actual value

    Raises
    ------
    TypeError
        Arguments must be integers or floats

    Returns
    -------
    residual : float
        Difference between the actual value and the expected value

    See Also
    --------
    :func:`~regressions.statistics.deviations.single_deviation`, :func:`~regressions.statistics.correlation.correlation_coefficient`

    Notes
    -----
    - Observed value: :math:`y`
    - Predicted value: :math:`\\hat{y}`
    - Residual: :math:`e = y - \\hat{y}`
    - |residual|

    Examples
    --------
    Import `single_residual` function from `regressions` library
        >>> from regressions.statistics.residuals import single_residual
    Determine the residual between an actual value of 7.8 and an expected value of 9.2
        >>> residual_small = single_residual(7.8, 9.2)
        >>> print(residual_small)
        -1.3999999999999995
    Determine the residual between an actual value of 6.1 and an expected value of 19.8
        >>> residual_large = single_residual(6.1, 19.8)
        >>> print(residual_large)
        -13.700000000000001
    """
    # Handle input errors
    scalar_value(actual, 'first')
    scalar_value(expected, 'second')

    # Calculate difference between inputs
    difference = actual - expected

    # Convert difference to float
    result = float(difference)
    return result

def multiple_residuals(actual_array, expected_array):
    """
    Generates a list of the differences between the actual values from one list and the expected values from another list

    Parameters
    ----------
    actual_array : list of int or float
        List containing the actual values observed from a data set
    expected_array : list of int or float
        List containing the expected values predicted for a data set

    Raises
    ------
    TypeError
        Arguments must be 1-dimensional lists
    TypeError
        Elements of arguments must be integers or floats
    ValueError
        Both arguments must contain the same number of elements

    Returns
    -------
    residuals : list of float
        Differences between the actual values and the expected values

    See Also
    --------
    :func:`~regressions.statistics.deviations.multiple_deviations`, :func:`~regressions.statistics.correlation.correlation_coefficient`

    Notes
    -----
    - Observed values: :math:`y_i = \\{ y_1, y_2, \\cdots, y_n \\}`
    - Predicted values: :math:`\\hat{y}_i = \\{ \\hat{y}_1, \\hat{y}_2, \\cdots, \\hat{y}_n \\}`
    - Residuals: :math:`e_i = \\{ y_1 - \\hat{y}_1, y_2 - \\hat{y}_2, \\cdots, y_n - \\hat{y}_n \\}`
    - |residual|

    Examples
    --------
    Import `multiple_residuals` function from `regressions` library
        >>> from regressions.statistics.residuals import multiple_residuals
    Determine the residuals between the actual values [5.6, 8.1, 6.3] and the expected values [6.03, 8.92, 6.12]
        >>> residuals_short = multiple_residuals([5.6, 8.1, 6.3], [6.03, 8.92, 6.12])
        >>> print(residuals_short)
        [-0.4300000000000006, -0.8200000000000003, 0.17999999999999972]
    Determine the residuals between the actual values [11.7, 5.6, 8.1, 13.4, 6.3] and the expected values [15.17, 6.03, 8.92, 9.42, 6.12]
        >>> residuals_long = multiple_residuals([11.7, 5.6, 8.1, 13.4, 6.3], [15.17, 6.03, 8.92, 9.42, 6.12])
        >>> print(residuals_long)
        [-3.4700000000000006, -0.4300000000000006, -0.8200000000000003, 3.9800000000000004, 0.17999999999999972]
    """
    # Handle input errors
    compare_vectors(actual_array, expected_array)

    # Create list to return
    results = []

    # Iterate over inputs
    for i in range(len(actual_array)):
        # Store residuals of corresponding elements in list to return
        results.append(single_residual(actual_array[i], expected_array[i]))
    
    # Return results
    return results


from regressions.errors.vectors import vector_of_scalars
from .summation import sum_value

def mean_value(data):
    """
    Determines the arithmetic mean of a data set

    Parameters
    ----------
    data : list of int or float
        List of numbers to analyze

    Raises
    ------
    TypeError
        Argument must be a 1-dimensional list
    TypeError
        Elements of argument must be integers or floats

    Returns
    -------
    mean : float
        Arithmetic mean of the data set

    See Also
    --------
    :func:`~regressions.statistics.summation.sum_value`, :func:`~regressions.statistics.median.median_value`, :func:`~regressions.statistics.correlation.correlation_coefficient`

    Notes
    -----
    - Set of numbers: :math:`a_i = \\{ a_1, a_2, \\cdots, a_n \\}`
    - Mean of all numbers in set: :math:`\\bar{a} = \\frac{1}{n}\\cdot{\\sum\\limits_{i=1}^n a_i}`
    - |mean|

    Examples
    --------
    Import `mean_value` function from `regressions` library
        >>> from regressions.statistics.mean import mean_value
    Determine the mean of the set [21, 53, 3, 68, 43, 9, 72, 19, 20, 1]
        >>> mean_even = mean_value([21, 53, 3, 68, 43, 9, 72, 19, 20, 1])
        >>> print(mean_even)
        30.9
    Determine the mean of the set [12, 81, 13, 8, 42, 72, 91, 20, 20]
        >>> mean_odd = mean_value([12, 81, 13, 8, 42, 72, 91, 20, 20])
        >>> print(mean_odd)
        39.888888888888886
    """
    # Handle input errors
    vector_of_scalars(data)

    # Calculate sum of all elements in input
    sum_of_data = sum_value(data)

    # Determine length of input
    length_of_data = len(data)

    # Calculate average
    average = sum_of_data / length_of_data

    # Convert average to float
    result = float(average)
    return result